@@ -7,7 +7,11 @@
 #include <stdint.h>
 
 #ifndef DQLITE_API
-#define DQLITE_API __attribute__((visibility("default")))
+# if defined(__has_attribute) && __has_attribute(visibility)
+#  define DQLITE_API __attribute__((visibility("default")))
+# else
+#  define DQLITE_API
+# endif
 #endif
 
 /**
@@ -28,13 +32,18 @@
  */
 #define DQLITE_VERSION_MAJOR 1
 #define DQLITE_VERSION_MINOR 18
-#define DQLITE_VERSION_RELEASE 0
+#define DQLITE_VERSION_RELEASE 3
 #define DQLITE_VERSION_NUMBER                                            \
 	(DQLITE_VERSION_MAJOR * 100 * 100 + DQLITE_VERSION_MINOR * 100 + \
 	 DQLITE_VERSION_RELEASE)
 
+#define SQLITE_IOERR_NOT_LEADER      (SQLITE_IOERR | (40 << 8))
+#define SQLITE_IOERR_LEADERSHIP_LOST (SQLITE_IOERR | (41 << 8))
+
 DQLITE_API int dqlite_version_number(void);
 
+DQLITE_API extern const char *dqlite_version_string;
+
 /**
  * Hold the value of a dqlite node ID. Guaranteed to be at least 64-bit long.
  */
@@ -215,7 +224,8 @@ DQLITE_API DQLITE_EXPERIMENTAL void dqlite_server_destroy(
  * These are used only with the dqlite_node family of functions.
  */
 enum {
-	DQLITE_ERROR = 1, /* Generic error */
+	DQLITE_OK = 0,
+	DQLITE_ERROR, /* Generic error */
 	DQLITE_MISUSE,    /* Library used incorrectly */
 	DQLITE_NOMEM      /* A malloc() failed */
 };
@@ -370,6 +380,37 @@ DQLITE_API int dqlite_node_set_network_latency_ms(dqlite_node *t,
 DQLITE_API int dqlite_node_set_failure_domain(dqlite_node *n,
 					      unsigned long long code);
 
+enum {
+	DQLITE_SNAPSHOT_TRAILING_STATIC = 0,
+	DQLITE_SNAPSHOT_TRAILING_DYNAMIC = 1,
+};
+
+/**
+ * !!! Deprecated, use `dqlite_node_set_snapshot_params_v2` instead which also includes
+ * trailing computation strategy. !!!
+ *
+ * Set the snapshot parameters for this node.
+ *
+ * This function determines how frequently a node will snapshot the state
+ * of the database and how many raft log entries will be kept around after
+ * a snapshot has been taken.
+ *
+ * `snapshot_threshold` : Determines the frequency of taking a snapshot, the
+ * lower the number, the higher the frequency.
+ *
+ * `snapshot_trailing` : Determines the maximum amount of log entries kept around after
+ * taking a snapshot. Lowering this number decreases disk and memory footprint
+ * but increases the chance of having to send a full snapshot (instead of a
+ * number of log entries to a node that has fallen behind).
+ *
+ * By default this function uses static trailing computation.
+ *
+ * This function must be called before calling dqlite_node_start().
+ */
+DQLITE_API int dqlite_node_set_snapshot_params(dqlite_node *n,
+					       unsigned snapshot_threshold,
+					       unsigned snapshot_trailing);
+
 /**
  * Set the snapshot parameters for this node.
  *
@@ -380,16 +421,30 @@ DQLITE_API int dqlite_node_set_failure_domain(dqlite_node *n,
  * `snapshot_threshold` : Determines the frequency of taking a snapshot, the
  * lower the number, the higher the frequency.
  *
- * `snapshot_trailing` : Determines the amount of log entries kept around after
+ * `snapshot_trailing` : Determines the maximum amount of log entries kept around after
  * taking a snapshot. Lowering this number decreases disk and memory footprint
  * but increases the chance of having to send a full snapshot (instead of a
- * number of log entries to a node that has fallen behind.
+ * number of log entries to a node that has fallen behind).
+ *
+ * `trailing_strategy` : Determines the strategy used to compute the number of
+ * trailing entries to keep after a snapshot has been taken. Valid values are
+ * `DQLITE_SNAPSHOT_TRAILING_STATIC` and `DQLITE_SNAPSHOT_TRAILING_DYNAMIC`.
+ *
+ * `DQLITE_SNAPSHOT_TRAILING_STATIC` will use directly the value of `snapshot_trailing`
+ * as the number of entries to keep after a snapshot has been taken.
+ *
+ * `DQLITE_SNAPSHOT_TRAILING_DYNAMIC` will compute the number of entries to keep
+ * by comparing the size of the snapshot to the size of the entries. The idea behind
+ * this is that if the amount of memory (on-disk or RAM) needed to store log entities
+ * exceeds the amount of memory for snapshot, streaming the snapshot is more efficient.
+ * The amount of entries kept is still capped at `snapshot_trailing`.
  *
  * This function must be called before calling dqlite_node_start().
  */
-DQLITE_API int dqlite_node_set_snapshot_params(dqlite_node *n,
+DQLITE_API int dqlite_node_set_snapshot_params_v2(dqlite_node *n,
 					       unsigned snapshot_threshold,
-					       unsigned snapshot_trailing);
+					       unsigned snapshot_trailing,
+					       int      trailing_strategy);
 
 /**
  * Set the block size used for performing disk IO when writing raft log segments
@@ -399,16 +454,6 @@ DQLITE_API int dqlite_node_set_snapshot_params(dqlite_node *n,
  */
 DQLITE_API int dqlite_node_set_block_size(dqlite_node *n, size_t size);
 
-/**
- * WARNING: This is an experimental API.
- *
- * By default dqlite holds the SQLite database file and WAL in memory. By
- * enabling disk-mode, dqlite will hold the SQLite database file on-disk while
- * keeping the WAL in memory. Has to be called after `dqlite_node_create` and
- * before `dqlite_node_start`.
- */
-DQLITE_API int dqlite_node_enable_disk_mode(dqlite_node *n);
-
 /**
  * Set the target number of voting nodes for the cluster.
  *
@@ -473,6 +518,14 @@ DQLITE_API int dqlite_node_set_snapshot_compression(dqlite_node *n,
  */
 DQLITE_API int dqlite_node_enable_role_management(dqlite_node *n);
 
+/**
+ * Set the amount of time in milliseconds a write query can stay in the write
+ * queue before failing with SQLITE_BUSY.
+ *
+ * This is 0ms by default to keep backward compatibility.
+ */
+DQLITE_API int dqlite_node_set_busy_timeout(dqlite_node *n, unsigned msecs);
+
 /**
  * Start a dqlite node.
  *
@@ -619,90 +672,6 @@ DQLITE_API const char *dqlite_node_errmsg(dqlite_node *n);
  */
 DQLITE_API dqlite_node_id dqlite_generate_node_id(const char *address);
 
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Initialize the given SQLite VFS interface object with dqlite's custom
- * implementation, which can be used for replication.
- */
-DQLITE_API int dqlite_vfs_init(sqlite3_vfs *vfs, const char *name);
-
-DQLITE_API int dqlite_vfs_enable_disk(sqlite3_vfs *vfs);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Release all memory used internally by a SQLite VFS object that was
- * initialized using @qlite_vfs_init.
- */
-DQLITE_API void dqlite_vfs_close(sqlite3_vfs *vfs);
-
-/**
- * This type is DEPRECATED and will be removed in a future major release.
- *
- * A single WAL frame to be replicated.
- */
-struct dqlite_vfs_frame
-{
-	unsigned long page_number; /* Database page number. */
-	void *data;                /* Content of the database page. */
-};
-typedef struct dqlite_vfs_frame dqlite_vfs_frame;
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Check if the last call to sqlite3_step() has triggered a write transaction on
- * the database with the given filename. In that case acquire a WAL write lock
- * to prevent further write transactions, and return all new WAL frames
- * generated by the transaction. These frames are meant to be replicated across
- * nodes and then actually added to the WAL with dqlite_vfs_apply() once a
- * quorum is reached. If a quorum is not reached within a given time, then
- * dqlite_vfs_abort() can be used to abort and release the WAL write lock.
- */
-DQLITE_API int dqlite_vfs_poll(sqlite3_vfs *vfs,
-			       const char *filename,
-			       dqlite_vfs_frame **frames,
-			       unsigned *n);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Add to the WAL all frames that were generated by a write transaction
- * triggered by sqlite3_step() and that were obtained via dqlite_vfs_poll().
- *
- * This interface is designed to match the typical use case of a node receiving
- * the frames by sequentially reading a byte stream from a network socket and
- * passing the data to this routine directly without any copy or futher
- * allocation, possibly except for integer encoding/decoding.
- */
-DQLITE_API int dqlite_vfs_apply(sqlite3_vfs *vfs,
-				const char *filename,
-				unsigned n,
-				unsigned long *page_numbers,
-				void *frames);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Abort a pending write transaction that was triggered by sqlite3_step() and
- * whose frames were obtained via dqlite_vfs_poll().
- *
- * This should be called if the transaction could not be safely replicated. In
- * particular it will release the write lock acquired by dqlite_vfs_poll().
- */
-DQLITE_API int dqlite_vfs_abort(sqlite3_vfs *vfs, const char *filename);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Return a snapshot of the main database file and of the WAL file.
- */
-DQLITE_API int dqlite_vfs_snapshot(sqlite3_vfs *vfs,
-				   const char *filename,
-				   void **data,
-				   size_t *n);
-
 /**
  * This type is DEPRECATED and will be removed in a future major release.
  *
@@ -714,53 +683,4 @@ struct dqlite_buffer
 	size_t len; /* Length of the buffer. */
 };
 
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Return a shallow snapshot of the main database file and of the WAL file.
- * Expects a bufs array of size x + 1, where x is obtained from
- * `dqlite_vfs_num_pages`.
- */
-DQLITE_API int dqlite_vfs_shallow_snapshot(sqlite3_vfs *vfs,
-					   const char *filename,
-					   struct dqlite_buffer bufs[],
-					   unsigned n);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- */
-DQLITE_API int dqlite_vfs_snapshot_disk(sqlite3_vfs *vfs,
-					const char *filename,
-					struct dqlite_buffer bufs[],
-					unsigned n);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Return the number of database pages (excluding WAL).
- */
-DQLITE_API int dqlite_vfs_num_pages(sqlite3_vfs *vfs,
-				    const char *filename,
-				    unsigned *n);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Restore a snapshot of the main database file and of the WAL file.
- */
-DQLITE_API int dqlite_vfs_restore(sqlite3_vfs *vfs,
-				  const char *filename,
-				  const void *data,
-				  size_t n);
-
-/**
- * This function is DEPRECATED and will be removed in a future major release.
- *
- * Restore a snapshot of the main database file and of the WAL file.
- */
-DQLITE_API int dqlite_vfs_restore_disk(sqlite3_vfs *vfs,
-				       const char *filename,
-				       const void *data,
-				       size_t main_size,
-				       size_t wal_size);
 #endif /* DQLITE_H */
