@@ -1,12 +1,17 @@
 #include "conn.h"
+#include "gateway.h"
+#include "leader.h"
 #include "message.h"
 #include "protocol.h"
 #include "request.h"
 #include "tracing.h"
 #include "transport.h"
+#include "utils.h"
 
 #include <uv.h>
 
+#define conn_trace(C, fmt, ...) tracef("[conn %p] "fmt, C, ##__VA_ARGS__)
+
 /* Initialize the given buffer for reading, ensure it has the given size. */
 static int init_read(struct conn *c, uv_buf_t *buf, size_t size)
 {
@@ -26,7 +31,7 @@ static void conn_write_cb(struct transport *transport, int status)
 	bool finished;
 	int rv;
 	if (status != 0) {
-		tracef("write cb status %d", status);
+		conn_trace(c, "write cb status %d", status);
 		goto abort;
 	}
 
@@ -56,6 +61,7 @@ static void gateway_handle_cb(struct handle *req,
 			      uint8_t type,
 			      uint8_t schema)
 {
+	(void)status;
 	struct conn *c = req->data;
 	size_t n;
 	char *cursor;
@@ -64,18 +70,12 @@ static void gateway_handle_cb(struct handle *req,
 
 	assert(schema <= req->schema);
 
-	/* Ignore results firing after we started closing. TODO: instead, we
-	 * should make gateway__close() asynchronous. */
+	/* Ignore results firing after we started closing. */
 	if (c->closed) {
-		tracef("gateway handle cb closed");
+		conn_trace(c, "gateway handle cb closed");
 		return;
 	}
 
-	if (status != 0) {
-		tracef("gateway handle cb status %d", status);
-		goto abort;
-	}
-
 	n = buffer__offset(&c->write) - message__sizeof(&c->response);
 	assert(n % 8 == 0);
 
@@ -92,15 +92,12 @@ static void gateway_handle_cb(struct handle *req,
 
 	rv = transport__write(&c->transport, &buf, conn_write_cb);
 	if (rv != 0) {
-		tracef("transport write failed %d", rv);
-		goto abort;
+		conn_trace(c, "transport write failed %d", rv);
+		conn__stop(c);
 	}
-	return;
-abort:
-	conn__stop(c);
 }
 
-static void closeCb(struct transport *transport)
+static void transportCloseCb(struct transport *transport)
 {
 	struct conn *c = transport->data;
 	buffer__close(&c->write);
@@ -115,10 +112,10 @@ static void raft_connect(struct conn *c)
 	struct cursor *cursor = &c->handle.cursor;
 	struct request_connect request;
 	int rv;
-	tracef("raft_connect");
+	conn_trace(c, "raft_connect");
 	rv = request_connect__decode(cursor, &request);
 	if (rv != 0) {
-		tracef("request connect decode failed %d", rv);
+		conn_trace(c, "request connect decode failed %d", rv);
 		conn__stop(c);
 		return;
 	}
@@ -127,7 +124,7 @@ static void raft_connect(struct conn *c)
 	/* Close the connection without actually closing the transport, since
 	 * the stream will be used by raft */
 	c->closed = true;
-	closeCb(&c->transport);
+	transportCloseCb(&c->transport);
 }
 
 static void read_request_cb(struct transport *transport, int status)
@@ -137,7 +134,7 @@ static void read_request_cb(struct transport *transport, int status)
 	int rv;
 
 	if (status != 0) {
-		tracef("read error %d", status);
+		conn_trace(c, "read error %d", status);
 		// errorf(c->logger, "read error");
 		conn__stop(c);
 		return;
@@ -158,7 +155,7 @@ static void read_request_cb(struct transport *transport, int status)
 	rv = gateway__handle(&c->gateway, &c->handle, c->request.type,
 			     c->request.schema, &c->write, gateway_handle_cb);
 	if (rv != 0) {
-		tracef("read gateway handle error %d", rv);
+		conn_trace(c, "read gateway handle error %d", rv);
 		conn__stop(c);
 	}
 }
@@ -173,7 +170,7 @@ static int read_request(struct conn *c)
 	}
 	rv = init_read(c, &buf, c->request.words * 8);
 	if (rv != 0) {
-		tracef("init read failed %d", rv);
+		conn_trace(c, "init read failed %d", rv);
 		return rv;
 	}
 	if (c->request.words == 0) {
@@ -181,7 +178,7 @@ static int read_request(struct conn *c)
 	}
 	rv = transport__read(&c->transport, &buf, read_request_cb);
 	if (rv != 0) {
-		tracef("transport read failed %d", rv);
+		conn_trace(c, "transport read failed %d", rv);
 		return rv;
 	}
 	return 0;
@@ -195,7 +192,7 @@ static void read_message_cb(struct transport *transport, int status)
 
 	if (status != 0) {
 		// errorf(c->logger, "read error");
-		tracef("read error %d", status);
+		conn_trace(c, "read error %d", status);
 		conn__stop(c);
 		return;
 	}
@@ -208,7 +205,7 @@ static void read_message_cb(struct transport *transport, int status)
 
 	rv = read_request(c);
 	if (rv != 0) {
-		tracef("read request error %d", rv);
+		conn_trace(c, "read request error %d", rv);
 		conn__stop(c);
 		return;
 	}
@@ -221,12 +218,12 @@ static int read_message(struct conn *c)
 	int rv;
 	rv = init_read(c, &buf, message__sizeof(&c->request));
 	if (rv != 0) {
-		tracef("init read failed %d", rv);
+		conn_trace(c, "init read failed %d", rv);
 		return rv;
 	}
 	rv = transport__read(&c->transport, &buf, read_message_cb);
 	if (rv != 0) {
-		tracef("transport read failed %d", rv);
+		conn_trace(c, "transport read failed %d", rv);
 		return rv;
 	}
 	return 0;
@@ -240,7 +237,7 @@ static void read_protocol_cb(struct transport *transport, int status)
 
 	if (status != 0) {
 		// errorf(c->logger, "read error");
-		tracef("read error %d", status);
+		conn_trace(c, "read error %d", status);
 		goto abort;
 	}
 
@@ -256,7 +253,7 @@ static void read_protocol_cb(struct transport *transport, int status)
 		/* c->protocol); */
 		/* TODO: instead of closing the connection we should return
 		 * error messages */
-		tracef("unknown protocol version %" PRIu64, c->protocol);
+		conn_trace(c, "unknown protocol version %" PRIu64, c->protocol);
 		goto abort;
 	}
 	c->gateway.protocol = c->protocol;
@@ -278,12 +275,12 @@ static int read_protocol(struct conn *c)
 	int rv;
 	rv = init_read(c, &buf, sizeof c->protocol);
 	if (rv != 0) {
-		tracef("init read failed %d", rv);
+		conn_trace(c, "init read failed %d", rv);
 		return rv;
 	}
 	rv = transport__read(&c->transport, &buf, read_protocol_cb);
 	if (rv != 0) {
-		tracef("transport read failed %d", rv);
+		conn_trace(c, "transport read failed %d", rv);
 		return rv;
 	}
 	return 0;
@@ -300,10 +297,10 @@ int conn__start(struct conn *c,
 {
 	int rv;
 	(void)loop;
-	tracef("conn start");
+	conn_trace(c, "start");
 	rv = transport__init(&c->transport, stream);
 	if (rv != 0) {
-		tracef("conn start - transport init failed %d", rv);
+		conn_trace(c, "start: transport init failed %d", rv);
 		goto err;
 	}
 	c->config = config;
@@ -319,7 +316,9 @@ int conn__start(struct conn *c,
 	if (rv != 0) {
 		goto err_after_read_buffer_init;
 	}
-	c->handle.data = c;
+	c->handle = (struct handle) {
+		.data = c,
+	};
 	c->closed = false;
 	/* First, we expect the client to send us the protocol version. */
 	rv = read_protocol(c);
@@ -338,13 +337,25 @@ err:
 	return rv;
 }
 
+static void gatewayCloseCb(struct gateway *g) {
+	struct conn *c = CONTAINER_OF(g, struct conn, gateway);
+	transport__close(&c->transport, transportCloseCb);
+}
+
 void conn__stop(struct conn *c)
 {
-	tracef("conn stop");
 	if (c->closed) {
+		conn_trace(c, "stop: already stopped");
 		return;
 	}
+	conn_trace(c, "stop");
 	c->closed = true;
-	gateway__close(&c->gateway);
-	transport__close(&c->transport, closeCb);
+	gateway__close(&c->gateway, gatewayCloseCb);
+}
+
+void conn__on_leadership_lost(struct conn *c)
+{
+	if (!c->closed) {
+		gateway__close_leader(&c->gateway);
+	}
 }
