@@ -9,19 +9,19 @@
 #include "lib/queue.h"
 
 #include "config.h"
+#include "raft.h"
 
 struct db
 {
 	struct config *config;        /* Dqlite configuration */
 	struct sqlite3_vfs *vfs;      /* Underlying VFS */
 	char *filename;               /* Database filename */
-	char *path;                   /* Used for on-disk db */
 	uint32_t cookie;              /* Used to bind to the pool's thread */
+	raft_index read_index;        /* Raft index to linearize reads */
 	int leaders;                  /* Open leader connections */
 	struct leader *active_leader; /* Current leader writing to the database */
 	queue pending_queue;          /* Queue of pending execs, used by leader */
 	queue queue;                  /* Prev/next database, used by the registry */
-	int read_lock;                /* Lock used by snapshots & checkpoints */
 };
 
 /**
