@@ -2,7 +2,9 @@
 #include <sqlite3.h>
 
 #include "bind.h"
+#include "db.h"
 #include "leader.h"
+#include "lib/registry.h"
 #include "lib/threadpool.h"
 #include "protocol.h"
 #include "query.h"
@@ -16,7 +18,7 @@
 
 #define RAFT_GATEWAY_PARSE 0xff01 /* Internal use only */
 
-static bool sqlite3_statement_empty(sqlite3 *conn, const char *sql)
+static bool is_statement_empty(sqlite3 *conn, const char *sql)
 {
 	if (sql == NULL || sql[0] == '\0') {
 		return true;
@@ -56,18 +58,25 @@ static void gateway__leader_close_cb(struct leader *leader)
 	}
 }
 
-static void gateway_finalize(struct gateway *g)
+void gateway__close_leader(struct gateway *g)
 {
 	if (g->leader != NULL) {
 		/* Before closing the gateway, signal to the existing leader that we
 		 * are closing and wait to drain the queue. */
 		leader__close(g->leader, gateway__leader_close_cb);
-	} else if (g->close_cb != NULL) {
+	} else {
 		stmt__registry_close(&g->stmts);
-		g->close_cb(g);
+		if (g->close_cb != NULL) {
+			g->close_cb(g);
+		}
 	}
 }
 
+static void gateway_finalize(struct gateway *g)
+{
+	gateway__close_leader(g);
+}
+
 void gateway__close(struct gateway *g, gateway_close_cb cb)
 {
 	PRE(cb != NULL);
@@ -312,7 +321,7 @@ static int handle_open(struct gateway *g, struct handle *req)
 			"a database for this connection is already open");
 		return 0;
 	}
-	rc = registry__db_get(g->registry, request.filename, &db);
+	rc = registry__get_or_create(g->registry, request.filename, &db);
 	if (rc != 0) {
 		tracef("registry db get failed %d", rc);
 		return rc;
@@ -364,7 +373,7 @@ static void handle_prepare_done_cb(struct exec *exec)
 
 
 	if (req->schema == DQLITE_PREPARE_STMT_SCHEMA_V0) {
-		if (!sqlite3_statement_empty(g->leader->conn, tail)) {
+		if (!is_statement_empty(g->leader->conn, tail)) {
 			sqlite3_finalize(stmt);
 			return failure(req, SQLITE_ERROR,
 				"nonempty statement tail");
@@ -474,14 +483,13 @@ static void fill_result(struct gateway *g, struct response_result *response)
 	response->rows_affected = (uint64_t)sqlite3_changes(g->leader->conn);
 }
 
-static void handle_exec_work_cb(struct exec *exec)
+static int exec_work(struct raft_io_async_work *work)
 {
-	PRE(exec->stmt != NULL);
+	struct exec *exec = work->data;
 	struct gateway *g = exec->data;
 	struct handle *req = g->req;
 
-	int rv = 0;
-	rv = bind__params(exec->stmt, &req->decoder);
+	int rv = bind__params(exec->stmt, &req->decoder);
 	if (rv != DQLITE_OK) {
 		leader_exec_result(exec, RAFT_GATEWAY_PARSE);
 	} else {
@@ -491,7 +499,28 @@ static void handle_exec_work_cb(struct exec *exec)
 				   rv == SQLITE_DONE ? RAFT_OK : RAFT_ERROR);
 	}
 
-	TAIL return leader_exec_resume(exec);
+	return RAFT_OK;
+}
+
+static void exec_work_done(struct raft_io_async_work *req, int status)
+{
+	assert(status == RAFT_OK);
+	return leader_exec_resume(req->data);
+}
+
+static void handle_exec_work_cb(struct exec *exec)
+{
+	PRE(exec->stmt != NULL);
+	struct gateway *g = exec->data;
+	g->work = (struct raft_io_async_work){
+		.data = exec,
+		.work = exec_work,
+	};
+	int rv = g->raft->io->async_work(g->raft->io, &g->work, exec_work_done);
+	if (rv != RAFT_OK) {
+		leader_exec_result(exec, rv);
+		return leader_exec_resume(exec);
+	}
 }
 
 static void handle_exec_done_cb(struct exec *exec)
@@ -655,39 +684,44 @@ static int handle_exec_sql(struct gateway *g, struct handle *req)
 	return 0;
 }
 
-static void handle_query_work_cb(struct exec *exec)
+static int query_work(struct raft_io_async_work *work)
 {
+	struct exec *exec = work->data;
+	struct gateway *g = exec->data;
+	struct handle *req = g->req;
+
+	int rv;
+	if (!req->parameters_bound) {
+		PRE(!sqlite3_stmt_busy(exec->stmt));
+		rv = bind__params(exec->stmt, &req->decoder);
+		if (rv != DQLITE_OK) {
+			leader_exec_result(exec, RAFT_GATEWAY_PARSE);
+			return SQLITE_ERROR;
+		}
+		if (tuple_decoder__remaining(&req->decoder) > 0) {
+			leader_exec_result(exec, RAFT_GATEWAY_PARSE);
+			return SQLITE_ERROR;
+		}
+		/* FIXME(marco6): Should I check if all bindings were consumed?
+		 * And moreover, should I allow parameters altogether in this case? */
+		req->parameters_bound = true;
+	}
+
+	return query__batch(exec->stmt, req->buffer);
+}
+
+static void query_work_done(struct raft_io_async_work *work, int rc)
+{
+	struct exec *exec = work->data;
 	struct gateway *g = exec->data;
-	PRE(g->req != NULL && g->leader != NULL);
 	struct handle *req = g->req;
 
 	if (req->cancellation_requested) {
 		/* Nothing else to do. */
-		TAIL return leader_exec_resume(exec);
+		sqlite3_reset(exec->stmt);
+		return leader_exec_resume(exec);
 	}
 
-	if (!sqlite3_statement_empty(exec->leader->conn, exec->tail)) {
-		leader_exec_result(exec, RAFT_ERROR);
-		TAIL return leader_exec_resume(exec);
-	}
-
-	exec->tail = NULL;
-	int rc;
-	if (!req->parameters_bound) {
-		PRE(!sqlite3_stmt_busy(exec->stmt));
-		rc = bind__params(exec->stmt, &req->decoder);
-		if (rc != DQLITE_OK ||
-		    tuple_decoder__remaining(&req->decoder) > 0) {
-			leader_exec_result(exec, RAFT_GATEWAY_PARSE);
-			TAIL return leader_exec_resume(exec);
-		}
-		/* FIXME(marco6): Should I check if all bindings were consumed?
-		 * And moreover, should I allow parameters altogether in this
-		 * case? */
-		req->parameters_bound = true;
-	}
-
-	rc = query__batch(exec->stmt, req->buffer);
 	if (rc == SQLITE_ROW) {
 		/* If the statement is still running, do not resume the
 		 * exec state machine, but send a response instead. The
@@ -700,8 +734,38 @@ static void handle_query_work_cb(struct exec *exec)
 		return;
 	}
 
-	leader_exec_result(exec, rc == SQLITE_DONE ? RAFT_OK : RAFT_ERROR);
-	TAIL return leader_exec_resume(exec);
+	leader_exec_result(exec, rc == SQLITE_DONE ? RAFT_OK : RAFT_ERROR);		
+	return leader_exec_resume(exec);
+}
+
+static void handle_query_work_cb(struct exec *exec)
+{
+	struct gateway *g = exec->data;
+	PRE(g->req != NULL && g->leader != NULL);
+	struct handle *req = g->req;
+
+	if (req->cancellation_requested) {
+		/* Nothing else to do. */
+		sqlite3_reset(exec->stmt);
+		return leader_exec_resume(exec);
+	}
+
+	if (!is_statement_empty(exec->leader->conn, exec->tail)) {
+		leader_exec_result(exec, RAFT_ERROR);
+		return leader_exec_resume(exec);
+	}
+
+	exec->tail = NULL;
+
+	g->work = (struct raft_io_async_work){
+		.data = exec,
+		.work = query_work,
+	};
+	int rv = g->raft->io->async_work(g->raft->io, &g->work, query_work_done);
+	if (rv != RAFT_OK) {
+		leader_exec_result(exec, rv);
+		return leader_exec_resume(exec);
+	}
 }
 
 static void handle_query_done_cb(struct exec *exec)
@@ -724,7 +788,6 @@ static void handle_query_done_cb(struct exec *exec)
 		goto done;
 	}
 
-
 	if (req->cancellation_requested) {
 		struct response_empty response = { 0 };
 		SUCCESS(empty, EMPTY, response, 0);
@@ -1033,104 +1096,80 @@ static int handle_remove(struct gateway *g, struct handle *req)
 }
 
 static int dumpFile(const char *filename,
-		    uint8_t *data,
-		    size_t n,
-		    struct buffer *buffer)
+		     const struct vfsSnapshot *file,
+		     struct buffer *buffer)
 {
 	char *cur;
-	uint64_t len = n;
+	uint64_t len = file->page_count * file->page_size;
 
 	cur = buffer__advance(buffer, text__sizeof(&filename));
 	if (cur == NULL) {
-		goto oom;
+		return DQLITE_NOMEM;
 	}
 	text__encode(&filename, &cur);
 	cur = buffer__advance(buffer, uint64__sizeof(&len));
 	if (cur == NULL) {
-		goto oom;
+		return DQLITE_NOMEM;
 	}
 	uint64__encode(&len, &cur);
 
-	if (n == 0) {
-		return 0;
+	if (len == 0) {
+		return DQLITE_OK;
 	}
 
-	assert(n % 8 == 0);
-	assert(data != NULL);
+	assert(len % 8 == 0);
 
-	cur = buffer__advance(buffer, n);
+	cur = buffer__advance(buffer, (size_t)len);
 	if (cur == NULL) {
-		goto oom;
+		return DQLITE_NOMEM;
 	}
-	memcpy(cur, data, n);
 
-	return 0;
+	for (size_t i = 0; i < file->page_count; i++) {
+		memcpy(cur, file->pages[i], file->page_size);
+		cur += file->page_size;
+	}
 
-oom:
-	return DQLITE_NOMEM;
+	return DQLITE_OK;
 }
 
 static int handle_dump(struct gateway *g, struct handle *req)
 {
 	tracef("handle dump");
 	struct cursor *cursor = &req->cursor;
-	bool err = true;
-	sqlite3_vfs *vfs;
 	char *cur;
 	char filename[1024] = { 0 };
-	void *data;
-	size_t n;
-	uint8_t *page;
-	uint32_t database_size = 0;
-	uint8_t *database;
-	uint8_t *wal;
-	size_t n_database;
-	size_t n_wal;
-	int rv;
 	START_V0(dump, files);
 
+	struct db *db = registry__get(g->registry, request.filename);
+	if (db == NULL) {
+		failure(req, DQLITE_NOTFOUND, "database does not exists");
+		return SQLITE_OK;
+	}
+
+	sqlite3 *conn;
+	int rv = db__open(db, &conn);
+	if (rv != SQLITE_OK) {
+		failure(req, rv, "failed to open database");
+		return DQLITE_OK;
+	}
+
+	struct vfsSnapshot snapshot;
+	rv = VfsAcquireSnapshot(conn, &snapshot);
+	if (rv != SQLITE_OK) {
+		failure(req, rv, "failed to open database");
+		sqlite3_close(conn);
+		return DQLITE_OK;
+	}
+
 	response.n = 2;
 	cur = buffer__advance(req->buffer, response_files__sizeof(&response));
 	assert(cur != NULL);
 	response_files__encode(&response, &cur);
 
-	/* It is not possible to use g->leader->db->vfs as dump call can
-	 * be issued without opening the leader connection first. */
-	vfs = sqlite3_vfs_find(g->config->name);
-	rv = VfsSnapshot(vfs, request.filename, &data, &n);
+	rv = dumpFile(request.filename, &snapshot, req->buffer);
 	if (rv != 0) {
-		tracef("dump failed");
-		failure(req, rv, "failed to dump database");
-		return 0;
-	}
-
-	if (data != NULL) {
-		/* Extract the database size from the first page. */
-		page = data;
-		database_size += (uint32_t)(page[28] << 24);
-		database_size += (uint32_t)(page[29] << 16);
-		database_size += (uint32_t)(page[30] << 8);
-		database_size += (uint32_t)(page[31]);
-
-		n_database = database_size * g->config->page_size;
-		n_wal = n - n_database;
-
-		database = data;
-		wal = database + n_database;
-	} else {
-		assert(n == 0);
-
-		n_database = 0;
-		n_wal = 0;
-
-		database = NULL;
-		wal = NULL;
-	}
-
-	rv = dumpFile(request.filename, database, n_database, req->buffer);
-	if (rv != 0) {
-		tracef("dump failed");
-		failure(req, rv, "failed to dump database");
+		tracef("main dump failed");
+		failure(req, rv, "failed to dump main database file");
 		goto out_free_data;
 	}
 
@@ -1142,25 +1181,20 @@ static int handle_dump(struct gateway *g, struct handle *req)
 	strncpy(filename, request.filename,
 		sizeof(filename) - strlen(wal_suffix) - 1);
 	strcat(filename, wal_suffix);
-	rv = dumpFile(filename, wal, n_wal, req->buffer);
+	static const struct vfsSnapshot empty_wal = {}; 
+	rv = dumpFile(filename, &empty_wal, req->buffer);
 	if (rv != 0) {
-		tracef("wal dump failed");
-		failure(req, rv, "failed to dump wal file");
+		tracef("WAL dump failed");
+		failure(req, rv, "failed to dump WAL file");
 		goto out_free_data;
 	}
 
-	err = false;
+	req->cb(req, 0, DQLITE_RESPONSE_FILES, 0);
 
 out_free_data:
-	if (data != NULL) {
-		raft_free(data);
-	}
-
-	if (!err) {
-		req->cb(req, 0, DQLITE_RESPONSE_FILES, 0);
-	}
-
-	return 0;
+	VfsReleaseSnapshot(conn, &snapshot);
+	sqlite3_close(conn);
+	return DQLITE_OK;
 }
 
 static int encodeServer(struct gateway *g,
@@ -1357,10 +1391,11 @@ int gateway__handle(struct gateway *g,
 handle:
 	req->type = type;
 	req->schema = schema;
-	req->cb = cb;
 	req->buffer = buffer;
 	req->db_id = 0;
+	req->cancellation_requested = false;
 	req->parameters_bound = 0;
+	req->cb = cb;
 	req->work = (pool_work_t){};
 
 	switch (type) {
