@@ -4,8 +4,6 @@
 #include <stdio.h>
 #include <time.h>
 
-#include "../include/dqlite.h"
-
 #include "command.h"
 #include "db.h"
 #include "leader.h"
@@ -24,19 +22,33 @@ static int exec_apply(struct exec *req,
 		      const struct vfsTransaction *transaction);
 static void exec_prepare_barrier_cb(struct raft_barrier *barrier, int status);
 static void exec_run_barrier_cb(struct raft_barrier *barrier, int status);
-static void exec_apply_cb(struct raft_apply *req, int status, void *result);
+static void exec_apply_cb(struct raft_apply *req, int status);
 static void exec_timer_cb(struct raft_timer *timer);
-static bool is_db_full(sqlite3_vfs *vfs, struct db *db, unsigned nframes);
+static bool is_db_full(sqlite3 *conn, unsigned nframes);
 
 static struct exec *exec_dequeue(struct db *db);
 static void exec_enqueue(struct db *db, struct exec *exec);
 
+static int progress_abort(void *db) {
+	(void)db;
+	return SQLITE_ABORT;
+}
+
 /* Whether we need to submit a barrier request because there is no transaction
  * in progress in the underlying database and the FSM is behind the last log
  * index. */
 static bool exec_needs_barrier(struct leader *l)
 {
-	return raft_last_applied(l->raft) < raft_last_index(l->raft);
+	if (sqlite3_txn_state(l->conn, NULL) != SQLITE_TXN_NONE) {
+		/* If a transaction is already in progress, there is little
+		 * benefit in submitting a barrier as the read mark is already
+		 * set. */
+		return false;
+	}
+	if (l->db->read_index == 0) {
+		l->db->read_index = raft_last_index(l->raft);
+	}
+	return l->db->read_index > raft_last_applied(l->raft);
 }
 
 int leader__init(struct leader *l, struct db *db, struct raft *raft)
@@ -64,119 +76,52 @@ static inline bool leader_closing(struct leader *leader)
 	return leader->close_cb != NULL;
 }
 
-static void leader_finalize(struct leader *leader)
+static struct exec *leader_finalize(struct leader *leader)
 {
 	PRE(leader->exec == NULL && leader->pending == 0);
 	PRE(leader->db->leaders > 0);
 	tracef("leader close");
-	sqlite3_interrupt(leader->conn);
-	int rc = sqlite3_close_v2(leader->conn);
-	assert(rc == 0);
+
+	leader->db->leaders--;
 	if (leader->db->active_leader == leader) {
 		leader_trace(leader, "done");
 		leader->db->active_leader = NULL;
 	}
-	leader->db->leaders--;
+
+	struct exec *next = exec_dequeue(leader->db);
+
+	sqlite3_progress_handler(leader->conn, 1, progress_abort, NULL);
+	int rc = sqlite3_close_v2(leader->conn);
+	assert(rc == 0);
+
 	leader->close_cb(leader);
+
+	return  next;
 }
 
 void leader__close(struct leader *leader, leader_close_cb close_cb)
 {
+	if (leader->close_cb != NULL) {
+		return;
+	}
+
 	leader->close_cb = close_cb;
 	if (leader->pending == 0) {
-		struct db *db = leader->db;
-		leader_finalize(leader);
-
-		struct exec *req = exec_dequeue(db);
+		struct exec *req = leader_finalize(leader);
 		if (req == NULL) {
 			return;
 		}
 
+		struct db *db = req->leader->db;
 		PRE(IN(db->active_leader, NULL, req->leader));
 		db->active_leader = req->leader;
 		return exec_tick(req);
+	} else {
+		/* Abort all queries as soon as possible. */
+		sqlite3_progress_handler(leader->conn, 1, progress_abort, NULL);
 	}
 }
 
-/* A checkpoint command that fails to commit is not a huge issue.
- * The WAL will not be checkpointed this time around on these nodes,
- * a new checkpoint command will be issued once the WAL on the leader reaches
- * threshold size again. It's improbable that the WAL in this way could grow
- * without bound, it would mean that apply frames commands commit without
- * issues, while the checkpoint command would somehow always fail to commit. */
-static void leaderCheckpointApplyCb(struct raft_apply *req,
-				    int status,
-				    void *result)
-{
-	(void)result;
-	raft_free(req);
-	if (status != 0) {
-		tracef("checkpoint apply failed %d", status);
-	}
-}
-
-/* Attempt to perform a checkpoint on nodes running a version of dqlite that
- * doens't perform autonomous checkpoints. For recent nodes, the checkpoint
- * command will just be a no-op.
- * This function will run after the WAL might have been checkpointed during a
- * call to `apply_frames`.
- * */
-static void leaderMaybeCheckpointLegacy(struct leader *leader)
-{
-	tracef("leader maybe checkpoint legacy");
-	struct sqlite3_file *wal = NULL;
-	struct raft_buffer buf;
-	struct command_checkpoint command;
-	sqlite3_int64 size;
-	int rv;
-
-	/* Get the database file associated with this connection */
-	rv = sqlite3_file_control(leader->conn, NULL,
-				  SQLITE_FCNTL_JOURNAL_POINTER, &wal);
-	assert(rv == SQLITE_OK); /* Should never fail */
-	if (wal == NULL || wal->pMethods == NULL) {
-		/* This might happen at the beginning of the leader life cycle, 
-		 * when no pages have been applied yet. */
-		return;
-	}
-	rv = wal->pMethods->xFileSize(wal, &size);
-	assert(rv == SQLITE_OK); /* Should never fail */
-
-	/* size of the WAL will be 0 if it has just been checkpointed on this
-	 * leader as a result of running apply_frames. */
-	if (size != 0) {
-		return;
-	}
-
-	tracef("issue checkpoint command");
-
-	/* Attempt to perfom a checkpoint across nodes that don't perform
-	 * autonomous snapshots. */
-	command.filename = leader->db->filename;
-	rv = command__encode(COMMAND_CHECKPOINT, &command, &buf);
-	if (rv != 0) {
-		tracef("encode failed %d", rv);
-		return;
-	}
-
-	struct raft_apply *apply = raft_malloc(sizeof(*apply));
-	if (apply == NULL) {
-		tracef("raft_malloc - no mem");
-		goto err_after_buf_alloc;
-	}
-	rv = raft_apply(leader->raft, apply, &buf, 1, leaderCheckpointApplyCb);
-	if (rv != 0) {
-		tracef("raft_apply failed %d", rv);
-		raft_free(apply);
-		goto err_after_buf_alloc;
-	}
-
-	return;
-
-err_after_buf_alloc:
-	raft_free(buf.base);
-}
-
 /**
  * State machine for exec requests.
  *
@@ -308,16 +253,17 @@ void leader_exec_abort(struct exec *req)
 	case EXEC_DONE: /* already done */
 		return;
 	case EXEC_RUNNING:
-		/* best-effort: there is no guarantee that this will interrupt the query */
-		sqlite3_interrupt(req->leader->conn);
+		/* This will abort only if the query is not already done. 
+		 * This will be reset when a new query is executed. */
+		sqlite3_progress_handler(req->leader->conn, 1, progress_abort, NULL);
 		return;
 	case EXEC_WAITING_QUEUE:
 		/* timers are cancellable, so the request can move on directly. */
 		leader_exec_result(req, RAFT_CANCELED);
-		TAIL return exec_tick(req);
+		return exec_tick(req);
 	}
 
-	/* Raft-related requests canno be cancelled, so the only step that can be taken
+	/* Raft-related requests cannot be cancelled, so the only step that can be taken
 	 * is to mark the request as failed and wait for the callback */
 	leader_exec_result(req, RAFT_CANCELED);
 }
@@ -341,7 +287,7 @@ void leader_exec_result(struct exec *req, int status)
 void leader_exec_resume(struct exec *req)
 {
 	PRE(sm_state(&req->sm) == EXEC_RUNNING);
-	TAIL return exec_tick(req);
+	return exec_tick(req);
 }
 
 static int exec_apply(struct exec *req, const struct vfsTransaction *transaction)
@@ -356,7 +302,7 @@ static int exec_apply(struct exec *req, const struct vfsTransaction *transaction
 	struct db *db = leader->db;
 	struct raft_buffer buf;
 
-	if (is_db_full(req->leader->db->vfs, req->leader->db, transaction->n_pages)) {
+	if (is_db_full(req->leader->conn, transaction->n_pages)) {
 		return SQLITE_FULL;
 	}
 
@@ -367,7 +313,7 @@ static int exec_apply(struct exec *req, const struct vfsTransaction *transaction
 		.is_commit = 1,
 		.frames = {
 			.n_pages = (uint32_t)transaction->n_pages,
-			.page_size = (uint16_t)db->config->page_size,
+			.page_size = (uint16_t)db->config->vfs.page_size,
 			.page_numbers = transaction->page_numbers,
 			.pages = transaction->pages,
 		}
@@ -450,11 +396,10 @@ static void exec_tick(struct exec *req)
 	PRE(req->leader != NULL && req->leader->db != NULL);
 	struct leader *leader = req->leader;
 	struct db *db = leader->db;
-	struct vfsTransaction transaction;
 
+	leader_trace(leader, "exec resume %s (status = %d)",
+		     exec_state_name(sm_state(&req->sm)), req->status);
 	for (;;) {
-		leader_trace(leader, "exec tick %s (status = %d)",
-			     exec_state_name(sm_state(&req->sm)), req->status);
 		switch (sm_state(&req->sm)) {
 		case EXEC_INITED:
 			PRE(leader->exec == NULL);
@@ -466,6 +411,10 @@ static void exec_tick(struct exec *req)
 				continue;
 			}
 
+			/* Unset the abort handler that might have been set to abort the
+			 * previous statement. */
+			sqlite3_progress_handler(req->leader->conn, 0, NULL, NULL);
+
 			if (req->stmt != NULL) {
 				sm_move(&req->sm, EXEC_PREPARED);
 				continue;
@@ -571,44 +520,50 @@ static void exec_tick(struct exec *req)
 
 			leader_trace(leader, "executing query");
 			sm_move(&req->sm, EXEC_RUNNING);
-			TAIL return req->work_cb(req);
-		case EXEC_RUNNING: /* -> EXEC_DONE */
+			return req->work_cb(req);
+		case EXEC_RUNNING:
 			leader_trace(leader, "executed query on leader (status=%d)", req->status);
 			if (req->status != RAFT_OK) {
 				sm_move(&req->sm, EXEC_DONE);
 				continue;
-			}
+			} else {
+				struct vfsTransaction transaction = {};
+				int rc = VfsPoll(leader->conn, &transaction);
+				if (rc != SQLITE_OK) {
+					leader_trace(leader,
+						     "poll failed on leader");
+					req->status = RAFT_IOERR;
+					sm_move(&req->sm, EXEC_WAITING_APPLY);
+					continue;
+				}
 
-			int rc = VfsPoll(leader->conn, &transaction);
-			if (rc != SQLITE_OK) {
-				leader_trace(leader, "poll failed on leader");
-				rc = VfsAbort(leader->conn);
-				assert(rc == SQLITE_OK);
-				req->status = RAFT_IOERR;
-				sm_move(&req->sm, EXEC_DONE);
+				leader_trace(leader,
+					     "polled connection (%d frames)",
+					     transaction.n_pages);
+				if (transaction.n_pages == 0) {
+					sm_move(&req->sm, EXEC_DONE);
+					continue;
+				}
+
+				req->status = exec_apply(req, &transaction);
+				for (unsigned i = 0; i < transaction.n_pages;
+				     i++) {
+					sqlite3_free(transaction.pages[i]);
+				}
+				sqlite3_free(transaction.pages);
+				sqlite3_free(transaction.page_numbers);
+
+				sm_move(&req->sm, EXEC_WAITING_APPLY);
+				if (req->status == 0) {
+					suspend;
+				}
 				continue;
 			}
-
-			leader_trace(leader, "polled connection (%d frames)", transaction.n_pages);
-			if (transaction.n_pages == 0) {
-				sm_move(&req->sm, EXEC_DONE);
-				continue;
-			}
-
-			req->status = exec_apply(req, &transaction);
-			for (unsigned i = 0; i < transaction.n_pages; i++) {
-				sqlite3_free(transaction.pages[i]);
-			}
-			sqlite3_free(transaction.pages);
-			sqlite3_free(transaction.page_numbers);
-
-			if (req->status != 0) {
-				sm_move(&req->sm, EXEC_DONE);
-				continue;
-			}
-			sm_move(&req->sm, EXEC_WAITING_APPLY);
-			suspend;
 		case EXEC_WAITING_APPLY:
+			if (req->status != RAFT_OK) {
+				int rv = VfsAbort(leader->conn);
+				assert(rv == SQLITE_OK);
+			}
 			sm_move(&req->sm, EXEC_DONE);
 			continue;
 		case EXEC_DONE: 
@@ -635,14 +590,15 @@ static void exec_tick(struct exec *req)
 			}
 
 			if (leader_closing(leader) && leader->pending == 0) {
-				leader_finalize(leader);
+				req = leader_finalize(leader);
+			} else {
+				req = exec_dequeue(db);
 			}
 
-			req = exec_dequeue(db);
 			if (req != NULL) {
 				PRE(IN(db->active_leader, NULL, req->leader));
 				db->active_leader = req->leader;
-				TAIL return exec_tick(req);
+				return exec_tick(req);
 			}
 			return;
 		default:
@@ -679,18 +635,14 @@ static void exec_timer_cb(struct raft_timer *timer)
 	return exec_tick(req);
 }
 
-static void exec_apply_cb(struct raft_apply *apply, int status, void *result)
+static void exec_apply_cb(struct raft_apply *apply, int status)
 {
-	(void)result;
 	struct exec *req = CONTAINER_OF(apply, struct exec, apply);
 	struct leader *leader = req->leader;
+	PRE(leader != NULL);
 	leader_trace(leader, "query applied (status=%d)", status);
-	if (leader) {
-		if (status != 0) {
-			VfsAbort(leader->conn);
-		} else {
-			leaderMaybeCheckpointLegacy(leader);
-		}
+	if (status == RAFT_OK) {
+		leader->db->read_index = apply->index;
 	}
 
 	PRE(sm_state(&req->sm) == EXEC_WAITING_APPLY);
@@ -699,8 +651,8 @@ static void exec_apply_cb(struct raft_apply *apply, int status, void *result)
 	return exec_tick(req);
 }
 
-static bool is_db_full(sqlite3_vfs *vfs, struct db *db, unsigned nframes)
+static bool is_db_full(sqlite3 *conn, unsigned nframes)
 {
-	uint64_t size = VfsDatabaseSize(vfs, db->path, nframes, db->config->page_size);
-	return size > VfsDatabaseSizeLimit(vfs);
+	uint64_t size = VfsDatabaseSize(conn, nframes);
+	return size > VfsDatabaseSizeLimit(conn);
 }
