@@ -36,7 +36,8 @@ int raft_version_number(void);
  * Error codes.
  */
 enum {
-	RAFT_NOMEM = 1,        /* Out of memory */
+	RAFT_OK = 0,
+	RAFT_NOMEM,            /* Out of memory */
 	RAFT_BADID,            /* Server ID is not valid */
 	RAFT_DUPLICATEID,      /* Server ID already in use */
 	RAFT_DUPLICATEADDRESS, /* Server address already in use */
@@ -58,7 +59,8 @@ enum {
 	RAFT_INVALID,      /* Invalid parameter */
 	RAFT_UNAUTHORIZED, /* No access to a resource */
 	RAFT_NOSPACE,      /* Not enough space on disk */
-	RAFT_TOOMANY       /* Some system or raft limit was hit */
+	RAFT_TOOMANY,      /* Some system or raft limit was hit */
+	RAFT_ERROR,        /* Generic error */
 };
 
 /**
@@ -307,10 +309,10 @@ struct page_from_to {
 	pageno_t to;
 };
 
-enum raft_result {
-	RAFT_RESULT_OK = 0,
-	RAFT_RESULT_UNEXPECTED = 1,
-	RAFT_RESULT_DONE = 2,
+enum raft_snapshot_result {
+	RAFT_SNAPSHOT_OK = 0,
+	RAFT_SNAPSHOT_UNEXPECTED = 1,
+	RAFT_SNAPSHOT_DONE = 2,
 };
 
 /**
@@ -325,14 +327,14 @@ struct raft_install_snapshot
 	struct raft_configuration conf; /* Config as of last_index. */
 	raft_index conf_index;          /* Commit index of conf. */
 	struct raft_buffer data;        /* Raw snapshot data. */
-	enum raft_result result;
+	enum raft_snapshot_result result;
 };
 #define RAFT_INSTALL_SNAPSHOT_VERSION 0
 
 struct raft_install_snapshot_result {
 	int version;
 
-	enum raft_result result;
+	enum raft_snapshot_result result;
 };
 #define RAFT_INSTALL_SNAPSHOT_RESULT_VERSION 0
 
@@ -342,7 +344,7 @@ struct raft_signature {
 	const char *db;
 	struct page_from_to page_from_to;
 	pageno_t cs_page_no;
-	enum raft_result result;
+	enum raft_snapshot_result result;
 	bool ask_calculated;
 };
 #define RAFT_SIGNATURE_VERSION 0
@@ -354,7 +356,7 @@ struct raft_signature_result {
 	struct page_checksum *cs;
 	unsigned int cs_nr;
 	pageno_t cs_page_no;
-	enum raft_result result;
+	enum raft_snapshot_result result;
 	bool calculated;
 };
 #define RAFT_SIGNATURE_RESULT_VERSION 0
@@ -365,7 +367,7 @@ struct raft_install_snapshot_mv {
 	const char *db;
 	struct page_from_to *mv;
 	unsigned int mv_nr;
-	enum raft_result result;
+	enum raft_snapshot_result result;
 };
 #define RAFT_INSTALL_SNAPSHOT_MV_VERSION 0
 
@@ -374,7 +376,7 @@ struct raft_install_snapshot_mv_result {
 
 	const char *db;
 	pageno_t last_known_page_no; /* used for retries and message losses */
-	enum raft_result result;
+	enum raft_snapshot_result result;
 };
 #define RAFT_INSTALL_SNAPSHOT_MV_RESULT_VERSION 0
 
@@ -384,7 +386,7 @@ struct raft_install_snapshot_cp {
 	const char *db;
 	pageno_t page_no;
 	struct raft_buffer page_data;
-	enum raft_result result;
+	enum raft_snapshot_result result;
 };
 #define RAFT_INSTALL_SNAPSHOT_CP_VERSION 0
 
@@ -392,7 +394,7 @@ struct raft_install_snapshot_cp_result {
 	int version;
 
 	pageno_t last_known_page_no; /* used for retries and message losses */
-	enum raft_result result;
+	enum raft_snapshot_result result;
 };
 #define RAFT_INSTALL_SNAPSHOT_CP_RESULT_VERSION 0
 
@@ -623,6 +625,15 @@ typedef void (*raft_io_recv_cb)(struct raft_io *io, struct raft_message *msg);
 
 typedef void (*raft_io_close_cb)(struct raft_io *io);
 
+struct raft_timer;
+typedef void (*raft_timer_cb)(struct raft_timer*);
+
+struct raft_timer {
+	void *data; /* User data */
+	raft_timer_cb cb; /* Request callback */
+	void *handle; /* Implementation handle */
+};
+
 /**
  * version field MUST be filled out by user.
  * When moving to a new version, the user MUST implement the newly added
@@ -630,7 +641,7 @@ typedef void (*raft_io_close_cb)(struct raft_io *io);
  */
 struct raft_io
 {
-	int version; /* 1 or 2 */
+	int version; /* 1 or 2 or 3 */
 	void *data;
 	void *impl;
 	char errmsg[RAFT_ERRMSG_BUF_SIZE];
@@ -683,8 +694,15 @@ struct raft_io
 	int (*random)(struct raft_io *io, int min, int max);
 	/* Field(s) below added since version 2. */
 	int (*async_work)(struct raft_io *io,
-			  struct raft_io_async_work *req,
-			  raft_io_async_work_cb cb);
+				struct raft_io_async_work *req,
+				raft_io_async_work_cb cb);
+	/* Field(s) below added since version 3. */
+	int (*timer_start)(struct raft_io *io,
+				struct raft_timer *req,
+				uint64_t timeout, uint64_t repeat,
+				raft_timer_cb cb);
+	int (*timer_stop)(struct raft_io *io,
+				struct raft_timer *req);
 };
 
 /**
@@ -720,16 +738,15 @@ struct raft_fsm
 	int version; /* 1, 2 or 3 */
 	void *data;
 	int (*apply)(struct raft_fsm *fsm,
-		     const struct raft_buffer *buf,
-		     void **result);
+		     const struct raft_buffer *buf);
 	int (*snapshot)(struct raft_fsm *fsm,
 			struct raft_buffer *bufs[],
 			unsigned *n_bufs);
 	int (*restore)(struct raft_fsm *fsm, struct raft_buffer *buf);
 	/* Fields below added since version 2. */
 	int (*snapshot_finalize)(struct raft_fsm *fsm,
-				 struct raft_buffer *bufs[],
-				 unsigned *n_bufs);
+				 struct raft_buffer bufs[],
+				 unsigned n_bufs);
 	/* Fields below added since version 3. */
 	int (*snapshot_async)(struct raft_fsm *fsm,
 			      struct raft_buffer *bufs[],
@@ -765,6 +782,29 @@ struct raft_transfer; /* Forward declaration */
 
 struct raft_log;
 
+/**
+ * Strategy to compute the amount of trailing entries to keep.
+ * This can be tuned with raft_set_snapshot_trailing_strategy
+ */
+enum {
+	/**
+	 * The static strategy is to just keep a fixed amount of entries
+	 * in the log cache. The number of entries can be tuned with 
+	 * @raft_set_snapshot_trailing. This is the default.
+	 */
+	RAFT_TRAILING_STRATEGY_STATIC,
+	
+	/**
+	 * The dynamic strategy computes the number of entries to keep by 
+	 * comparing the size held by each entry to the size of the snapshot.
+	 * The idea behind this is that if the amount of memory (on-disk or RAM)
+	 * needed to store log entities exceeds the amount of memory for snapshot,
+	 * streaming the snapshot is more efficient. The number of entries is still
+	 * limited by the value set by @raft_set_snapshot_trailing.
+	 */
+	RAFT_TRAILING_STRATEGY_DYNAMIC,
+};
+
 /**
  * Hold and drive the state of a single raft server in a cluster.
  * When replacing reserved fields in the middle of this struct, you MUST use a
@@ -964,7 +1004,10 @@ struct raft
 		unsigned trailing;  /* N. of trailing entries to retain */
 		struct raft_snapshot pending;    /* In progress snapshot */
 		struct raft_io_snapshot_put put; /* Store snapshot request */
-		uint64_t reserved[8];            /* Future use */
+		uint8_t trailing_strategy;
+		/* Future use */
+		uint8_t reserved2[7];
+		uint64_t reserved[7];
 	} snapshot;
 
 	/*
@@ -1107,12 +1150,17 @@ RAFT_API void raft_set_snapshot_threshold(struct raft *r, unsigned n);
 RAFT_API void raft_set_pre_vote(struct raft *r, bool enabled);
 
 /**
- * Number of outstanding log entries to keep in the log after a snapshot has
- * been taken. This avoids sending snapshots when a follower is behind by just a
- * few entries. The default is 128.
+ * Number of maximum outstanding log entries to keep in the log after a snapshot
+ * has been taken. This avoids sending snapshots when a follower is behind by just
+ * a few entries. The default is 2048.
  */
 RAFT_API void raft_set_snapshot_trailing(struct raft *r, unsigned n);
 
+/**
+ * Strategy to compute trailing amount. The default is RAFT_TRAILING_STRATEGY_STATIC.
+ */
+RAFT_API void raft_set_snapshot_trailing_strategy(struct raft *r, int strategy);
+
 /**
  * Set the maximum number of a catch-up rounds to try when replicating entries
  * to a stand-by server that is being promoted to voter, before giving up and
@@ -1158,6 +1206,11 @@ RAFT_API raft_index raft_last_index(struct raft *r);
  */
 RAFT_API raft_index raft_last_applied(struct raft *r);
 
+/**
+ * Return the index of the last entry that was committed.
+ */
+RAFT_API raft_index raft_commit_index(const struct raft *r);
+
 /**
  * Return the number of voting servers that the leader has recently been in
  * contact with. This can be used to help determine whether the cluster may be
@@ -1194,7 +1247,7 @@ RAFT_API int raft_voter_contacts(struct raft *r);
  * the FSM when a quorum is reached.
  */
 struct raft_apply;
-typedef void (*raft_apply_cb)(struct raft_apply *req, int status, void *result);
+typedef void (*raft_apply_cb)(struct raft_apply *req, int status);
 struct raft_apply
 {
 	RAFT__REQUEST;
@@ -1239,6 +1292,10 @@ struct raft_barrier
 {
 	RAFT__REQUEST;
 	raft_barrier_cb cb;
+
+	/* Singly linked list for bundled requests targetting the same raft
+	 * index. */
+	struct raft_barrier *next;
 };
 
 /**
@@ -1326,6 +1383,21 @@ RAFT_API int raft_transfer(struct raft *r,
 			   raft_id id,
 			   raft_transfer_cb cb);
 
+/**
+ * Starts a timer.
+ */
+RAFT_API int raft_timer_start(struct raft *r,
+				struct raft_timer *req,
+				uint64_t timeout, uint64_t repeat,
+				raft_timer_cb cb);
+
+/**
+ * Stops a timer. If the timer wasn't started,
+ * this is a no-op.
+ */
+RAFT_API int raft_timer_stop(struct raft *r,
+				struct raft_timer *req);
+
 /**
  * User-definable dynamic memory allocation functions.
  *
@@ -1373,8 +1445,6 @@ DQLITE_VISIBLE_TO_TESTS void raft_heap_set_default(void);
  */
 DQLITE_VISIBLE_TO_TESTS const struct raft_heap *raft_heap_get(void);
 
-#undef RAFT__REQUEST
-
 struct raft_uv_transport;
 
 /**
@@ -1910,6 +1980,17 @@ RAFT_API bool raft_fixture_step_until_applied(struct raft_fixture *f,
 					      raft_index index,
 					      unsigned max_msecs);
 
+/**
+ * Step the cluster until the @i'th server has committed the entry at the given
+ * index, or @max_msecs have elapsed. If @i equals the number of servers, then
+ * step until all servers have committed the given entry.
+ */
+RAFT_API bool raft_fixture_step_until_committed(struct raft_fixture *f,
+				     unsigned i,
+				     raft_index index,
+				     unsigned max_msecs);
+
+
 /**
  * Step the cluster until the state of the @i'th server matches the given one,
  * or @max_msecs have elapsed.
@@ -2046,6 +2127,10 @@ RAFT_API void raft_fixture_set_send_latency(struct raft_fixture *f,
 					    unsigned j,
 					    unsigned msecs);
 
+RAFT_API void raft_fixture_set_work_duration(struct raft_fixture *f,
+					     unsigned i,
+					     unsigned msecs);
+
 /**
  * Set the persisted term of the @i'th server.
  */
