@@ -1,8 +1,11 @@
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
+#include "../lib/queue.h"
+#include "../lib/sm.h" /* struct sm */
 #include "../raft.h"
 #include "../tracing.h"
 #include "assert.h"
@@ -10,7 +13,6 @@
 #include "convert.h"
 #include "entry.h"
 #include "log.h"
-#include "../lib/queue.h"
 #include "snapshot.h"
 
 /* Defaults */
@@ -58,6 +60,33 @@ RAFT_API unsigned raft_fixture_event_server_index(
 	return event->server_index;
 }
 
+/**
+ * State machine for I/O requests in the fixture.
+ */
+enum {
+	FIO_START,
+	FIO_END,
+	FIO_FAIL,
+	FIO_NR,
+};
+
+#define A(ident) BITS(FIO_##ident)
+#define S(ident, allowed_, flags_) \
+	[FIO_##ident] = { .name = #ident, .allowed = (allowed_), .flags = (flags_) }
+
+static const struct sm_conf fio_states[FIO_NR] = {
+	S(START, A(END)|A(FAIL), SM_INITIAL),
+	S(END,   0,              SM_FINAL),
+	S(FAIL,  0,              SM_FAILURE|SM_FINAL),
+};
+
+static bool fio_invariant(const struct sm *sm, int prev)
+{
+	(void)sm;
+	(void)prev;
+	return true;
+}
+
 /* Fields common across all request types. */
 #define REQUEST                                                                \
 	int type;                  /* Request code type. */                    \
@@ -65,7 +94,16 @@ RAFT_API unsigned raft_fixture_event_server_index(
 	queue queue                /* Link the I/O pending requests queue. */
 
 /* Request type codes. */
-enum { APPEND = 1, SEND, TRANSMIT, SNAPSHOT_PUT, SNAPSHOT_GET, ASYNC_WORK };
+enum {
+	APPEND = 1,
+	TRUNCATE,
+	SEND,
+	TRANSMIT,
+	SNAPSHOT_PUT,
+	SNAPSHOT_GET,
+	ASYNC_WORK,
+	TIMER,
+};
 
 /* Abstract base type for an asynchronous request submitted to the stub I/o
  * implementation. */
@@ -84,6 +122,12 @@ struct append
 	unsigned start; /* Request timestamp. */
 };
 
+struct truncate {
+	REQUEST;
+	struct raft_io_truncate *req;
+	raft_index index;
+};
+
 /* Pending request to send a message. */
 struct send
 {
@@ -124,6 +168,12 @@ struct transmit
 	int timer;                   /* Deliver after this n of msecs. */
 };
 
+struct timer {
+	REQUEST;
+	struct raft_timer *timer;
+	uint64_t repeat;
+};
+
 /* Information about a peer server. */
 struct peer
 {
@@ -228,6 +278,7 @@ static void ioFlushAppend(struct io *s, struct append *append)
 	/* Simulates a disk write failure. */
 	if (faultTick(&s->append_fault_countdown)) {
 		status = RAFT_IOERR;
+		sm_fail(&append->req->sm, FIO_FAIL, status);
 		goto done;
 	}
 
@@ -247,13 +298,42 @@ static void ioFlushAppend(struct io *s, struct append *append)
 	s->entries = entries;
 	s->n += append->n;
 
+	sm_move(&append->req->sm, FIO_END);
+
 done:
+	sm_fini(&append->req->sm);
 	if (append->req->cb != NULL) {
 		append->req->cb(append->req, status);
 	}
 	raft_free(append);
 }
 
+static void ioFlushTruncate(struct io *io, struct truncate *r)
+{
+	size_t n = (size_t)(r->index - 1); /* Number of entries left after truncation */
+	struct raft_entry *entries;
+
+	PRE(n <= io->n);
+	PRE(ERGO(io->n > 0, io->entries != NULL));
+	for (size_t i = n; i < io->n; i++) {
+		raft_free(io->entries[i].buf.base);
+	}
+	if (n == 0) {
+		raft_free(io->entries);
+		entries = NULL;
+	} else {
+		entries = raft_realloc(io->entries, n * sizeof(*io->entries));
+		POST(entries != NULL);
+	}
+
+	io->entries = entries;
+	io->n = n;
+	sm_move(&r->req->sm, FIO_END);
+	sm_fini(&r->req->sm);
+	raft_free(r->req);
+	raft_free(r);
+}
+
 /* Flush a snapshot put request, copying the snapshot data. */
 static void ioFlushSnapshotPut(struct io *s, struct snapshot_put *r)
 {
@@ -270,7 +350,9 @@ static void ioFlushSnapshotPut(struct io *s, struct snapshot_put *r)
 	assert(rv == 0);
 
 	if (r->trailing == 0) {
-		rv = s->io->truncate(s->io, 1);
+		struct raft_io_truncate *trunc = raft_malloc(sizeof(*trunc));
+		assert(trunc != NULL);
+		rv = s->io->truncate(s->io, trunc, 1);
 		assert(rv == 0);
 	}
 
@@ -304,6 +386,15 @@ static void ioFlushAsyncWork(struct io *s, struct async_work *r)
 	raft_free(r);
 }
 
+static void ioFlushTimer(struct io *s, struct timer *r)
+{
+	if (r->repeat) {
+		r->completion_time = *s->time + r->repeat;
+		queue_insert_tail(&s->requests, &r->queue);
+	}
+	r->timer->cb(r->timer);
+}
+
 /* Search for the peer with the given ID. */
 static struct peer *ioGetPeer(struct io *io, raft_id id)
 {
@@ -428,6 +519,9 @@ static void ioFlushAll(struct io *io)
 			case APPEND:
 				ioFlushAppend(io, (struct append *)r);
 				break;
+			case TRUNCATE:
+				ioFlushTruncate(io, (struct truncate *)r);
+				break;
 			case SEND:
 				ioFlushSend(io, (struct send *)r);
 				break;
@@ -445,6 +539,9 @@ static void ioFlushAll(struct io *io)
 			case ASYNC_WORK:
 				ioFlushAsyncWork(io, (struct async_work *)r);
 				break;
+			case TIMER:
+				ioFlushTimer(io, (struct timer *)r);
+				break;
 			default:
 				assert(0);
 		}
@@ -584,6 +681,8 @@ static int ioMethodAppend(struct raft_io *raft_io,
 	r = raft_malloc(sizeof *r);
 	assert(r != NULL);
 
+	sm_init(&req->sm, fio_invariant, NULL, fio_states, "fio-append", FIO_START);
+
 	r->type = APPEND;
 	r->completion_time = *io->time + io->disk_latency;
 	r->req = req;
@@ -597,48 +696,25 @@ static int ioMethodAppend(struct raft_io *raft_io,
 	return 0;
 }
 
-static int ioMethodTruncate(struct raft_io *raft_io, raft_index index)
+static int ioMethodTruncate(struct raft_io *raft_io,
+			    struct raft_io_truncate *trunc,
+			    raft_index index)
 {
 	struct io *io = raft_io->impl;
-	size_t n;
-
-	n = (size_t)(index - 1); /* Number of entries left after truncation */
-
-	if (n > 0) {
-		struct raft_entry *entries;
-
-		/* Create a new array of entries holding the non-truncated
-		 * entries */
-		entries = raft_malloc(n * sizeof *entries);
-		if (entries == NULL) {
-			return RAFT_NOMEM;
-		}
-		memcpy(entries, io->entries, n * sizeof *io->entries);
-
-		/* Release any truncated entry */
-		if (io->entries != NULL) {
-			size_t i;
-			for (i = n; i < io->n; i++) {
-				raft_free(io->entries[i].buf.base);
-			}
-			raft_free(io->entries);
-		}
-		io->entries = entries;
-	} else {
-		/* Release everything we have */
-		if (io->entries != NULL) {
-			size_t i;
-			for (i = 0; i < io->n; i++) {
-				raft_free(io->entries[i].buf.base);
-			}
-			raft_free(io->entries);
-			io->entries = NULL;
-		}
-	}
-
-	io->n = n;
+	struct truncate *r;
 
+	sm_init(&trunc->sm, fio_invariant, NULL, fio_states, "fio-trunc", FIO_START);
+	r = raft_malloc(sizeof(*r));
+	assert(r != NULL);
+	*r = (struct truncate){
+		.type = TRUNCATE,
+		.completion_time = *io->time + io->disk_latency,
+		.req = trunc,
+		.index = index,
+	};
+	queue_insert_tail(&io->requests, &r->queue);
 	return 0;
+
 }
 
 static int ioMethodSnapshotPut(struct raft_io *raft_io,
@@ -723,6 +799,36 @@ static int ioMethodRandom(struct raft_io *raft_io, int min, int max)
 	}
 }
 
+int ioMethodTimerStart(struct raft_io *raft_io, struct raft_timer *req, uint64_t timeout, uint64_t repeat, raft_timer_cb cb)
+{
+	struct io *io = raft_io->impl;
+	struct timer *t;
+	t = raft_malloc(sizeof *t);
+	assert(t != NULL);
+	t->type = TIMER;
+	t->completion_time = *io->time + timeout;
+	t->timer = req;
+	t->repeat = repeat;
+
+	req->handle = t;
+	req->cb = cb;
+	queue_insert_tail(&io->requests, &t->queue);
+	return 0;
+}
+
+int ioMethodTimerStop(struct raft_io *raft_io, struct raft_timer *req)
+{
+	(void)raft_io;
+	struct timer *t = req->handle;
+	if (t == NULL) {
+		return 0;
+	}
+	queue_remove(&t->queue);
+	raft_free(t);
+	req->handle = NULL;
+	return 0;
+}
+
 /* Queue up a request which will be processed later, when io_stub_flush()
  * is invoked. */
 static int ioMethodSend(struct raft_io *raft_io,
@@ -897,7 +1003,7 @@ static int ioInit(struct raft_io *raft_io, unsigned index, raft_time *time)
 	io->n_append = 0;
 
 	raft_io->impl = io;
-	raft_io->version = 2;
+	raft_io->version = 3;
 	raft_io->init = ioMethodInit;
 	raft_io->close = ioMethodClose;
 	raft_io->start = ioMethodStart;
@@ -914,6 +1020,8 @@ static int ioInit(struct raft_io *raft_io, unsigned index, raft_time *time)
 	raft_io->snapshot_get = ioMethodSnapshotGet;
 	raft_io->time = ioMethodTime;
 	raft_io->random = ioMethodRandom;
+	raft_io->timer_start = ioMethodTimerStart;
+	raft_io->timer_stop = ioMethodTimerStop;
 
 	return 0;
 }
@@ -1317,7 +1425,7 @@ static void copyLeaderLog(struct raft_fixture *f)
 		assert(buf.base != NULL);
 		memcpy(buf.base, entry->buf.base, buf.len);
 		/* FIXME(cole) what to do here for is_local? */
-		rv = logAppend(f->log, entry->term, entry->type, buf, (struct raft_entry_local_data){}, false, NULL);
+		rv = logAppend(f->log, entry->term, entry->type, buf, false, NULL);
 		assert(rv == 0);
 	}
 	logRelease(raft->log, 1, entries, n);
@@ -1407,6 +1515,10 @@ static void completeRequest(struct raft_fixture *f, unsigned i, raft_time t)
 			ioFlushAppend(io, (struct append *)r);
 			f->event->type = RAFT_FIXTURE_DISK;
 			break;
+		case TRUNCATE:
+			ioFlushTruncate(io, (struct truncate *)r);
+			f->event->type = RAFT_FIXTURE_DISK;
+			break;
 		case SEND:
 			ioFlushSend(io, (struct send *)r);
 			f->event->type = RAFT_FIXTURE_NETWORK;
@@ -1427,6 +1539,10 @@ static void completeRequest(struct raft_fixture *f, unsigned i, raft_time t)
 			ioFlushAsyncWork(io, (struct async_work *)r);
 			f->event->type = RAFT_FIXTURE_WORK;
 			break;
+		case TIMER:
+			ioFlushTimer(io, (struct timer *)r);
+			f->event->type = RAFT_FIXTURE_WORK;
+			break;
 		default:
 			assert(0);
 	}
@@ -1681,6 +1797,36 @@ bool raft_fixture_step_until_applied(struct raft_fixture *f,
 	return raft_fixture_step_until(f, hasAppliedIndex, &apply, max_msecs);
 }
 
+static bool hasCommittedIndex(struct raft_fixture *f, void *arg)
+{
+	const struct step_apply *apply = (struct step_apply *)arg;
+	struct raft *raft;
+	unsigned n = 0;
+	unsigned i;
+
+	if (apply->i < f->n) {
+		raft = raft_fixture_get(f, apply->i);
+		return raft_commit_index(raft) >= apply->index;
+	}
+
+	for (i = 0; i < f->n; i++) {
+		raft = raft_fixture_get(f, i);
+		if (raft_commit_index(raft) >= apply->index) {
+			n++;
+		}
+	}
+	return n == f->n;
+}
+
+bool raft_fixture_step_until_committed(struct raft_fixture *f,
+				     unsigned i,
+				     raft_index index,
+				     unsigned max_msecs)
+{
+	struct step_apply apply = {i, index};
+	return raft_fixture_step_until(f, hasCommittedIndex, &apply, max_msecs);
+}
+
 struct step_state
 {
 	unsigned i;
@@ -1932,6 +2078,14 @@ void raft_fixture_set_term(struct raft_fixture *f, unsigned i, raft_term term)
 	io->term = term;
 }
 
+void raft_fixture_set_work_duration(struct raft_fixture *f,
+				    unsigned i,
+				    unsigned msecs)
+{
+	struct io *io = f->servers[i]->io.impl;
+	io->work_duration = msecs;
+}
+
 void raft_fixture_set_snapshot(struct raft_fixture *f,
 			       unsigned i,
 			       struct raft_snapshot *snapshot)
