@@ -511,7 +511,7 @@ static void appendLeaderCb(struct raft_io_append *append, int status)
 					struct raft_apply *apply =
 					    (struct raft_apply *)req;
 					if (apply->cb) {
-						apply->cb(apply, status, NULL);
+						apply->cb(apply, status);
 					}
 					break;
 				}
@@ -1247,11 +1247,14 @@ int replicationAppend(struct raft *r,
 	 *   entry).
 	 */
 	if (n == 0) {
+		PRE(r->last_applied <= r->commit_index);
 		if ((args->leader_commit > r->commit_index) &&
 		    r->last_stored >= r->commit_index &&
 		    !replicationInstallSnapshotBusy(r)) {
 			r->commit_index =
 			    min(args->leader_commit, r->last_stored);
+		}
+		if (r->last_applied < r->commit_index) {
 			rv = replicationApply(r);
 			if (rv != 0) {
 				goto err_after_request_alloc;
@@ -1553,10 +1556,9 @@ static int applyCommand(struct raft *r,
 			const struct raft_buffer *buf)
 {
 	struct raft_apply *req;
-	void *result;
 	int rv;
 
-	rv = r->fsm->apply(r->fsm, buf, &result);
+	rv = r->fsm->apply(r->fsm, buf);
 	if (rv != 0) {
 		return rv;
 	}
@@ -1571,7 +1573,7 @@ static int applyCommand(struct raft *r,
 	sm_move(&req->sm, REQUEST_COMPLETE);
 	sm_fini(&req->sm);
 	if (req->cb != NULL) {
-		req->cb(req, 0, result);
+		req->cb(req, 0);
 	}
 	return 0;
 }
@@ -1680,7 +1682,9 @@ static void takeSnapshotClose(struct raft *r, struct raft_snapshot *s)
 	}
 
 	configurationClose(&s->configuration);
-	r->fsm->snapshot_finalize(r->fsm, &s->bufs, &s->n_bufs);
+	r->fsm->snapshot_finalize(r->fsm, s->bufs, s->n_bufs);
+	s->bufs = NULL;
+	s->n_bufs = 0;
 }
 
 
@@ -1884,6 +1888,7 @@ int replicationApply(struct raft *r)
 		return 0;
 	}
 
+	tracef("Apply indexed %d...%d", r->last_applied, r->commit_index);
 	for (index = r->last_applied + 1; index <= r->commit_index; index++) {
 		const struct raft_entry *entry = logGet(r->log, index);
 		if (entry == NULL) {
@@ -1934,6 +1939,8 @@ int replicationApply(struct raft *r)
 
 	if (shouldTakeSnapshot(r)) {
 		rv = takeSnapshot(r);
+	} else if (rv == RAFT_BUSY) {
+		rv = RAFT_OK;
 	}
 
 	return rv;
