@@ -1,16 +1,34 @@
-#include "uv_fs.h"
+#ifndef _GNU_SOURCE
+# define _GNU_SOURCE
+#endif
 
+#include <fcntl.h>
+#include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/vfs.h>
+#include <time.h>
 #include <unistd.h>
 
 #include "assert.h"
-#include "compress.h"
 #include "err.h"
 #include "heap.h"
+#include "uv_fs.h"
 #include "uv_os.h"
 
+#ifdef HAVE_XFS_XFS_H
+# include <xfs/xfs.h> 
+# include <linux/magic.h> 
+#endif
+
+#ifdef LZ4_AVAILABLE
+# include <lz4.h>
+# include <lz4frame.h>
+#endif
+
+#define UV__FS_PROBE_FILE ".probe"
+#define UV__FS_PROBE_FILE_SIZE 4096
+
 int UvFsCheckDir(const char *dir, char *errmsg)
 {
 	struct uv_fs_s req;
@@ -145,6 +163,99 @@ int UvFsFileIsEmpty(const char *dir,
 	return 0;
 }
 
+/* Reopens a file descriptor with FMODE_NOCMTIME flag on for XFS.
+ *
+ * This is needed because XFS does not work correctly with async I/O when using
+ * O_DSYNC: unlike other filesystems, it also updates metadata (ctime/mtime)
+ * synchronously, which breaks async behavior. Disabling metadata updates allows
+ * dqlite to use fdatasync as intended.
+ *
+ * TODO: remove this hack if/when we get `O_CMTIME` flag.
+ */
+static int uvXfsMaybeReopen(int fd, const char *dir, int flags)
+{
+	/* Given that if this fails the worst thing happening is going back to
+	 * the thread pool, errors here are ignored and the reopening just
+	 * doesn't happen. This will likely be the case, given that (as of
+	 * Linux 6.17) SYS_ADMIN capability is required for this flow to suceed.
+	 *
+	 * As of Linux 6.17, there is no userland way to properly control this
+	 * behaviour, but there is a comment around FMODE_NOCMTIME that gives
+	 * some insight (linux/fs.h):
+	 *
+	 *    Don't update ctime and mtime.
+	 *
+	 *    Currently a special hack for the XFS open_by_handle ioctl, but
+	 *    we'll hopefully graduate it to a proper O_CMTIME flag supported
+	 *    by open(2) soon.
+	 *
+	 * Which points to the idea that it is possible to opt-out this
+	 * behaviour using the syscall `open_by_handle_at` that, as a hack, will
+	 * set the internal FMODE_NOCMTIME flag to suppress this behaviour.
+	 *
+	 * See the linux kernel functions:
+	 *  - xfs_file_write_checks
+	 *  - kiocb_modified
+	 *  - file_modified_flags
+	 * for a rational on the checks and:
+	 *  - xfs_open_by_handle
+	 * for the mentioned hack.
+	 */
+#ifdef HAVE_XFS_XFS_H
+	/* Only for synchronized I/O */
+	if (!(flags & O_DSYNC)) {
+		return fd;
+	}
+
+	/* Fix for xfs */
+	struct statfs fs_info;
+	if (fstatfs(fd, &fs_info) == -1) {
+		return fd;
+	}
+
+	if (fs_info.f_type != XFS_SUPER_MAGIC) {
+		return fd;
+	}
+
+	/* Open the containing directory. */
+	int dirfd = open(dir, O_RDONLY);
+	if (dirfd < 0) {
+		return fd;
+	}
+
+    xfs_handle_t handle = {};
+    __u32 handle_length = sizeof(handle);
+	xfs_fsop_handlereq_t req = {
+		.fd = (unsigned)fd,
+	    .ohandle = &handle,
+	    .ohandlen = &handle_length, 
+	};
+
+	int rv = ioctl(dirfd, XFS_IOC_FD_TO_HANDLE, &req);
+	if (rv != 0) {
+		close(dirfd);
+		return fd;
+	}
+	
+	req = (xfs_fsop_handlereq_t){
+	    .ihandle = &handle,
+	    .ihandlen = handle_length,
+        .oflags = (__u32)(flags & ~(O_CREAT | O_EXCL)),
+	};
+	int new_fd = ioctl(dirfd, XFS_IOC_OPEN_BY_HANDLE, &req);
+	close(dirfd);
+	if (new_fd < 0) {
+		return fd;
+	}
+	close(fd);
+	return new_fd;
+#else
+	(void)dir;
+	(void)flags;
+	return fd;
+#endif
+}
+
 /* Open a file in a directory. */
 static int uvFsOpenFile(const char *dir,
 			const char *filename,
@@ -164,6 +275,9 @@ static int uvFsOpenFile(const char *dir,
 		UvOsErrMsg(errmsg, "open", rv);
 		return RAFT_IOERR;
 	}
+
+	*fd = uvXfsMaybeReopen(*fd, dir, flags);
+
 	return 0;
 }
 
@@ -192,7 +306,9 @@ int UvFsAllocateFile(const char *dir,
 		     char *errmsg)
 {
 	char path[UV__PATH_SZ];
-	int flags = O_WRONLY | O_CREAT | O_EXCL; /* Common open flags */
+	/* TODO: use RWF_DSYNC instead, if available. */
+	const int create_flags = O_CREAT | O_EXCL;
+	const int open_flags = O_WRONLY | O_DSYNC; /* Common open flags */
 	int rv = 0;
 
 	rv = UvOsJoin(dir, filename, path);
@@ -202,10 +318,8 @@ int UvFsAllocateFile(const char *dir,
 
 	/* Allocate the desired size. */
 	if (fallocate) {
-		/* TODO: use RWF_DSYNC instead, if available. */
-		flags |= O_DSYNC;
-		rv = uvFsOpenFile(dir, filename, flags, S_IRUSR | S_IWUSR, fd,
-				  errmsg);
+		rv = uvFsOpenFile(dir, filename, open_flags | create_flags,
+				  S_IRUSR | S_IWUSR, fd, errmsg);
 		if (rv != 0) {
 			goto err;
 		}
@@ -226,8 +340,9 @@ int UvFsAllocateFile(const char *dir,
 	} else {
 		/* Emulate fallocate, open without O_DSYNC, because we risk
 		 * doing a lot of synced writes. */
-		rv = uvFsOpenFile(dir, filename, flags, S_IRUSR | S_IWUSR, fd,
-				  errmsg);
+		rv = uvFsOpenFile(dir, filename,
+				  (open_flags | create_flags) & ~O_DSYNC,
+				  S_IRUSR | S_IWUSR, fd, errmsg);
 		if (rv != 0) {
 			goto err;
 		}
@@ -256,15 +371,14 @@ int UvFsAllocateFile(const char *dir,
 			goto err_unlink;
 		}
 		/* TODO: use RWF_DSYNC instead, if available. */
-		flags = O_WRONLY | O_DSYNC;
-		rv = uvFsOpenFile(dir, filename, flags, S_IRUSR | S_IWUSR, fd,
-				  errmsg);
+		rv = uvFsOpenFile(dir, filename, open_flags, S_IRUSR | S_IWUSR,
+				  fd, errmsg);
 		if (rv != 0) {
 			goto err_unlink;
 		}
 	}
 
-	return 0;
+	return RAFT_OK;
 
 err_after_open:
 	UvOsClose(*fd);
@@ -275,6 +389,7 @@ err:
 	return rv;
 }
 
+
 static int uvFsWriteFile(const char *dir,
 			 const char *filename,
 			 int flags,
@@ -389,6 +504,181 @@ err_after_tmp_create:
 	return rv;
 }
 
+#ifdef LZ4_AVAILABLE
+static inline int uvOsWriteOne(uv_os_fd_t fd, void *data, size_t length, int64_t offset, char *errmsg) {
+	const uv_buf_t buffer = {
+		.base = data,
+		.len = length,
+	};
+	int rv = UvOsWrite(fd, &buffer, 1, offset);
+	if (rv == (int)length) {
+		return RAFT_OK;
+	} else if (rv < 0) {
+		UvOsErrMsg(errmsg, "write", rv);
+	} else {
+		ErrMsgPrintf(errmsg, "short write: %d only bytes written", rv);
+	}
+	return RAFT_IOERR;
+}
+#endif
+
+int UvFsMakeCompressedFile(const char *dir,
+			   const char *filename,
+			   struct raft_buffer *bufs,
+			   unsigned n_bufs,
+			   char *errmsg)
+{
+#ifndef LZ4_AVAILABLE
+	(void)dir;
+	(void)filename;
+	(void)bufs;
+	(void)n_bufs;
+
+	ErrMsgPrintf(errmsg, "LZ4 not available");
+	return RAFT_INVALID;
+#else
+	char path[UV__PATH_SZ] = {};
+	int rv = UvOsJoin(dir, TMP_FILE_PREFIX "XXXXXX", path);
+	if (rv != 0) {
+		return RAFT_INVALID;
+	}
+
+	uv_fs_t temp_file;
+	rv = uv_fs_mkstemp(NULL, &temp_file, path, NULL);
+	if (rv == -1) {
+		UvOsErrMsg(errmsg, "mkstemp", rv);
+		return RAFT_IOERR;
+	}
+	uv_file fd = (uv_file)temp_file.result;
+
+	size_t chunk_size = 0;
+	size_t content_size = 0;
+	for (unsigned i = 0; i < n_bufs; i++) {
+		if (bufs[i].len > chunk_size) {
+			chunk_size = bufs[i].len;
+		}
+		content_size += bufs[i].len;
+	}
+
+	/* Limit the size of each chunk to 4MB to make sure this logic doesn't
+	 * need too much memory. */
+	const size_t lz4_max_block_size = 4 * 1024 * 1024;
+	assert(chunk_size <= lz4_max_block_size);
+
+	LZ4F_preferences_t
+	    lz4_pref = { .frameInfo = {
+			     /* Detect data corruption when decompressing */
+			     .contentChecksumFlag = 1,
+			     /* For allocating a suitable buffer when
+				decompressing */
+			     .contentSize = content_size,
+			 } };
+
+	const size_t output_cap = LZ4F_compressBound(chunk_size, &lz4_pref);
+	char *output_buffer = raft_malloc(output_cap);
+	if (output_buffer == NULL) {
+		rv = RAFT_NOMEM;
+		goto err_after_open;
+	}
+
+	LZ4F_compressionContext_t ctx;
+	LZ4F_errorCode_t err =
+	    LZ4F_createCompressionContext(&ctx, LZ4F_VERSION);
+	if (LZ4F_isError(err)) {
+		ErrMsgPrintf(errmsg, "LZ4F_createDecompressionContext %s",
+			     LZ4F_getErrorName(err));
+		rv = RAFT_NOMEM;
+		goto err_after_buf_alloc;
+	}
+
+	size_t output_len =
+	    LZ4F_compressBegin(ctx, output_buffer, output_cap, &lz4_pref);
+	if (LZ4F_isError(output_len)) {
+		ErrMsgPrintf(errmsg, "LZ4F_compressBegin %s",
+			     LZ4F_getErrorName(output_len));
+		rv = RAFT_IOERR;
+		goto err_after_ctx_alloc;
+	}
+
+	/* Write the header */
+	rv = uvOsWriteOne(fd, output_buffer, output_len, -1, errmsg);
+	if (rv != RAFT_OK) {
+		goto err_after_ctx_alloc;
+	}
+
+	for (unsigned i = 0; i < n_bufs; i++) {
+		output_len =
+		    LZ4F_compressUpdate(ctx, output_buffer, output_cap,
+					bufs[i].base, bufs[i].len, NULL);
+		if (output_len == 0) {
+			/* In this case the output is buffered internally by
+			 * liblz4 */
+			continue;
+		} else if (LZ4F_isError(output_len)) {
+			ErrMsgPrintf(errmsg, "LZ4F_compressUpdate %s",
+				     LZ4F_getErrorName(output_len));
+			rv = RAFT_IOERR;
+			goto err_after_ctx_alloc;
+		}
+
+		rv = uvOsWriteOne(fd, output_buffer, output_len, -1, errmsg);
+		if (rv != RAFT_OK) {
+			goto err_after_ctx_alloc;
+		}
+	}
+
+	output_len = LZ4F_compressEnd(ctx, output_buffer, output_cap, NULL);
+	if (LZ4F_isError(output_len)) {
+		ErrMsgPrintf(errmsg, "LZ4F_compressEnd %s",
+			     LZ4F_getErrorName(output_len));
+		rv = RAFT_IOERR;
+		goto err_after_ctx_alloc;
+	} else if (output_len > 0) {
+		rv = uvOsWriteOne(fd, output_buffer, output_len, -1, errmsg);
+		if (rv != RAFT_OK) {
+			goto err_after_ctx_alloc;
+		}
+	}
+
+	rv = UvOsFsync(fd);
+	if (rv != 0) {
+		UvOsErrMsg(errmsg, "fsync", rv);
+		goto err_after_ctx_alloc;
+	}
+
+	rv = UvOsJoin(dir, filename, path);
+	if (rv != 0) {
+		rv = RAFT_INVALID;
+		goto err_after_ctx_alloc;
+	}
+
+	rv = UvOsRename(temp_file.path, path);
+	if (rv != 0) {
+		rv = RAFT_IOERR;
+		goto err_after_ctx_alloc;
+	}
+
+	rv = UvFsSyncDir(dir, errmsg);
+	if (rv != 0) {
+		/* Try to unlink the new file. */
+		UvOsUnlink(path);
+	}
+
+err_after_ctx_alloc:
+	LZ4F_freeCompressionContext(ctx);
+err_after_buf_alloc:
+	raft_free(output_buffer);
+err_after_open:
+	UvOsClose(fd);
+	if (rv != RAFT_OK) {
+		/* Try to unlink the temp file. */
+		UvOsUnlink(temp_file.path);
+	}
+	uv_fs_req_cleanup(&temp_file);
+	return rv;
+#endif
+}
+
 int UvFsMakeOrOverwriteFile(const char *dir,
 			    const char *filename,
 			    const struct raft_buffer *buf,
@@ -519,6 +809,9 @@ int UvFsReadFile(const char *dir,
 		goto err;
 	}
 
+	rv = posix_fadvise(fd, 0, 0, POSIX_FADV_SEQUENTIAL);
+	assert(rv == 0);
+
 	buf->len = (size_t)sb.st_size;
 	buf->base = RaftHeapMalloc(buf->len);
 	if (buf->base == NULL) {
@@ -544,6 +837,147 @@ err:
 	return rv;
 }
 
+int UvFsReadCompressedFile(const char *dir,
+			   const char *filename,
+			   struct raft_buffer *buf,
+			   char *errmsg)
+{
+#ifndef LZ4_AVAILABLE
+	(void)dir;
+	(void)filename;
+	(void)buf;
+
+	ErrMsgPrintf(errmsg, "LZ4 not available");
+	return RAFT_INVALID;
+#else
+	char path[UV__PATH_SZ] = {};
+	int rv = UvOsJoin(dir, filename, path);
+	if (rv != 0) {
+		return RAFT_INVALID;
+	}
+
+	uv_file fd;
+	rv = uvFsOpenFile(dir, filename, O_RDONLY, 0, &fd, errmsg);
+	if (rv != 0) {
+		return rv;
+	}
+
+	rv = posix_fadvise(fd, 0, 0, POSIX_FADV_SEQUENTIAL);
+	assert(rv == 0);
+
+	LZ4F_decompressionContext_t ctx;
+	size_t lzrv = LZ4F_createDecompressionContext(&ctx, LZ4F_VERSION);
+	if (LZ4F_isError(lzrv)) {
+		ErrMsgPrintf(errmsg, "LZ4F_createDecompressionContext %s",
+			LZ4F_getErrorName(lzrv));
+		rv = RAFT_NOMEM;
+		goto err_after_open;
+	}
+
+	/* The proper input size depends a lot on what the maximum block size is, but 
+	 * in general it is possible to be a little bit pessimistic here and use the
+	 * max value, which is 4MiB. */
+	const size_t input_buffer_size =  LZ4_COMPRESSBOUND(4 * 1024 * 1024);
+	void *input_buffer = raft_malloc(input_buffer_size);
+	if (input_buffer == NULL) {
+		rv = RAFT_NOMEM;
+		goto err_after_ctx_alloc;
+	}
+
+	/* Read at least LZ4F_HEADER_SIZE_MAX bytes. */
+	size_t input_size = 0;
+	while (input_size < LZ4F_HEADER_SIZE_MAX) {
+		ssize_t read_rv = read(fd, (char *)input_buffer + input_size, input_buffer_size - input_size);
+		if (read_rv == -1) {
+			rv = RAFT_IOERR;
+			UvOsErrMsg(errmsg, "read", -errno);
+			goto done;
+		} else if (read_rv == 0) {
+			break;
+		}
+		input_size += (size_t)read_rv;
+	}
+	if (input_size < LZ4F_HEADER_SIZE_MIN) {
+		ErrMsgPrintf(errmsg, "compressed file is too short");
+		rv = RAFT_INVALID;
+		goto done;
+	}
+
+	LZ4F_frameInfo_t frameInfo = {};
+	size_t input_offset = (size_t)input_size;
+	lzrv = LZ4F_getFrameInfo(ctx, &frameInfo, input_buffer, &input_offset);
+	if (LZ4F_isError(lzrv)) {
+		ErrMsgPrintf(errmsg, "LZ4F_getFrameInfo %s",
+			     LZ4F_getErrorName(lzrv));
+		rv = RAFT_IOERR;
+		goto done;
+	}
+
+	if (frameInfo.contentSize == 0) {
+		rv = RAFT_OK;
+		goto done; // Not really an error here...
+	}
+
+	const size_t output_buffer_size = (size_t)frameInfo.contentSize;
+	void *output_buffer = raft_malloc(output_buffer_size);
+	if (output_buffer == NULL) {
+		rv = RAFT_NOMEM;
+		goto done;
+	}
+
+	size_t output_offset = 0;
+	while (output_offset < output_buffer_size) {
+		if (input_offset == input_size) {
+			/* try to read some */
+			ssize_t read_rv = read(fd, input_buffer, input_buffer_size);
+			if (read_rv < 0) {
+				UvOsErrMsg(errmsg, "read", -errno);
+				rv = RAFT_IOERR;
+				goto err_after_output_alloc;
+			} else if (read_rv == 0) {
+				ErrMsgPrintf(errmsg, "short read: %zu bytes instead of %zu",
+			     	output_offset, output_buffer_size);
+				rv = RAFT_IOERR;
+				goto err_after_output_alloc;
+			}
+			input_size = (size_t)read_rv;
+			input_offset = 0;
+		}
+
+		size_t output_size = output_buffer_size - output_offset;
+		size_t input_read = input_size - input_offset;
+		lzrv = LZ4F_decompress(ctx, output_buffer + output_offset,
+				&output_size, input_buffer + input_offset,
+				&input_read, NULL);
+		if (LZ4F_isError(lzrv)) {
+			ErrMsgPrintf(errmsg, "LZ4F_decompress %s",
+					LZ4F_getErrorName(lzrv));
+			rv = RAFT_IOERR;
+			goto err_after_output_alloc;
+		}
+
+		output_offset += output_size;
+		input_offset += input_read;
+	}
+	rv = RAFT_OK;
+	*buf = (struct raft_buffer) {
+		.base = output_buffer,
+		.len = output_buffer_size,
+	};
+	goto done;
+
+err_after_output_alloc:
+	raft_free(output_buffer);
+done:
+	raft_free(input_buffer);
+err_after_ctx_alloc:
+	LZ4F_freeDecompressionContext(ctx);
+err_after_open:
+	UvOsClose(fd);
+	return rv;
+#endif
+}
+
 int UvFsReadFileInto(const char *dir,
 		     const char *filename,
 		     struct raft_buffer *buf,
@@ -712,7 +1146,7 @@ static int probeDirectIO(int fd, size_t *size, char *errmsg)
 	}
 
 	/* Try to perform direct I/O, using various buffer size. */
-	*size = 4096;
+	*size = UV__FS_PROBE_FILE_SIZE;
 	while (*size >= 512) {
 		buf = raft_aligned_alloc(*size, *size);
 		if (buf == NULL) {
@@ -758,9 +1192,6 @@ static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
 {
 	void *buf;                  /* Buffer to use for the probe write */
 	aio_context_t ctx = 0;      /* KAIO context handle */
-	struct iocb iocb;           /* KAIO request object */
-	struct iocb *iocbs = &iocb; /* Because the io_submit() API sucks */
-	struct io_event event;      /* KAIO response object */
 	int n_events;
 	int rv;
 
@@ -781,16 +1212,16 @@ static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
 	memset(buf, 0, size);
 
 	/* Prepare the KAIO request object */
-	memset(&iocb, 0, sizeof iocb);
-	iocb.aio_lio_opcode = IOCB_CMD_PWRITE;
-	*((void **)(&iocb.aio_buf)) = buf;
-	iocb.aio_nbytes = size;
-	iocb.aio_offset = 0;
-	iocb.aio_fildes = (uint32_t)fd;
-	iocb.aio_reqprio = 0;
-	iocb.aio_rw_flags |= RWF_NOWAIT | RWF_DSYNC;
+	struct iocb iocb = {
+		.aio_lio_opcode = IOCB_CMD_PWRITE,
+		.aio_fildes = (__u32)fd,
+		.aio_buf = (__u64)(uintptr_t)buf,
+		.aio_nbytes = (__u64)size,
+		.aio_rw_flags = RWF_NOWAIT,
+	};
 
 	/* Submit the KAIO request */
+	struct iocb *iocbs = &iocb; /* Because the io_submit() API sucks */
 	rv = UvOsIoSubmit(ctx, 1, &iocbs);
 	if (rv != 0) {
 		/* UNTESTED: in practice this should fail only with ENOMEM */
@@ -808,6 +1239,7 @@ static int probeAsyncIO(int fd, size_t size, bool *ok, char *errmsg)
 	}
 
 	/* Fetch the response: will block until done. */
+	struct io_event event = {}; /* KAIO response object */
 	n_events = UvOsIoGetevents(ctx, 1, 1, &event, NULL);
 	assert(n_events == 1);
 	if (n_events != 1) {
@@ -866,8 +1298,6 @@ out:
 	UvFsRemoveFile(dir, UV__FS_PROBE_FALLOCATE_FILE, ignored);
 }
 
-#define UV__FS_PROBE_FILE ".probe"
-#define UV__FS_PROBE_FILE_SIZE 4096
 int UvFsProbeCapabilities(const char *dir,
 			  size_t *direct,
 			  bool *async,
