@@ -14,10 +14,11 @@
 #include "fsm.h"
 #include "lib/addr.h"
 #include "lib/assert.h"
-#include "lib/fs.h"
+#include "lib/queue.h"
 #include "lib/threadpool.h"
 #include "logger.h"
 #include "protocol.h"
+#include "raft.h"
 #include "roles.h"
 #include "tracing.h"
 #include "translate.h"
@@ -37,8 +38,26 @@ static void state_cb(struct raft *r,
 		     unsigned short old_state,
 		     unsigned short new_state)
 {
+	struct dqlite_node *d = r->data;
 	if (old_state == RAFT_LEADER && new_state != RAFT_LEADER) {
 		tracef("node %llu@%s: leadership lost", r->id, r->address);
+		queue *head;
+		QUEUE_FOREACH(head, &d->conns)
+		{
+			struct conn *conn =
+			    QUEUE_DATA(head, struct conn, queue);
+			conn__on_leadership_lost(conn);
+		}
+	} else if (old_state != RAFT_LEADER && new_state == RAFT_LEADER) {
+		/* Invalidate read index for all databases, so that they will
+		 * wait for the leader to have all entries applied before
+		 * reading. */
+		queue *head;
+		QUEUE_FOREACH(head, &d->registry.dbs)
+		{
+			struct db *db = QUEUE_DATA(head, struct db, queue);
+			db->read_index = 0;
+		}
 	}
 }
 
@@ -48,26 +67,18 @@ int dqlite__init(struct dqlite_node *d,
 		 const char *dir)
 {
 	int rv;
-	char db_dir_path[1024];
 
 	d->initialized = false;
 	d->lock_fd = -1;
 	memset(d->errmsg, 0, sizeof d->errmsg);
 
-	rv = snprintf(db_dir_path, sizeof db_dir_path, DATABASE_DIR_FMT, dir);
-	if (rv == -1 || rv >= (int)(sizeof db_dir_path)) {
-		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE,
-			 "failed to init: snprintf(rv:%d)", rv);
-		goto err;
-	}
-
-	rv = config__init(&d->config, id, address, dir, db_dir_path);
+	rv = config__init(&d->config, id, address, dir);
 	if (rv != 0) {
 		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE,
 			 "config__init(rv:%d)", rv);
 		goto err;
 	}
-	rv = VfsInit(&d->vfs, d->config.name);
+	rv = VfsInit(&d->vfs, &d->config.vfs);
 	sqlite3_vfs_register(&d->vfs, 0);
 	if (rv != 0) {
 		goto err_after_config_init;
@@ -426,31 +437,6 @@ int dqlite_node_set_block_size(dqlite_node *n, size_t size)
 	return 0;
 }
 
-int dqlite_node_enable_disk_mode(dqlite_node *n)
-{
-	int rv;
-
-	if (n->running) {
-		return DQLITE_MISUSE;
-	}
-
-	rv = VfsEnableDisk(&n->vfs);
-	if (rv != 0) {
-		return rv;
-	}
-
-	n->registry.config->disk = true;
-
-	/* Close the default fsm and initialize the disk one. */
-	fsm__close(&n->raft_fsm);
-	rv = fsm__init_disk(&n->raft_fsm, &n->config, &n->registry);
-	if (rv != 0) {
-		return rv;
-	}
-
-	return 0;
-}
-
 static int maybeBootstrap(dqlite_node *d,
 			  dqlite_node_id id,
 			  const char *address)
@@ -839,31 +825,6 @@ static void release_dir(int fd)
 	close(fd);
 }
 
-static int dqliteDatabaseDirSetup(dqlite_node *t)
-{
-	int rv;
-	if (!t->config.disk) {
-		// nothing to do
-		return 0;
-	}
-
-	rv = FsEnsureDir(t->config.database_dir);
-	if (rv != 0) {
-		snprintf(t->errmsg, DQLITE_ERRMSG_BUF_SIZE,
-			 "Error creating database dir: %d", rv);
-		return rv;
-	}
-
-	rv = FsRemoveDirFiles(t->config.database_dir);
-	if (rv != 0) {
-		snprintf(t->errmsg, DQLITE_ERRMSG_BUF_SIZE,
-			 "Error removing files in database dir: %d", rv);
-		return rv;
-	}
-
-	return rv;
-}
-
 int dqlite_node_start(dqlite_node *t)
 {
 	int rv;
@@ -871,12 +832,6 @@ int dqlite_node_start(dqlite_node *t)
 
 	dqliteTracingMaybeEnable(true);
 
-	rv = dqliteDatabaseDirSetup(t);
-	if (rv != 0) {
-		tracef("database dir setup failed %s", t->errmsg);
-		goto err;
-	}
-
 	int lock_fd;
 	rv = acquire_dir(t->config.raft_dir, &lock_fd);
 	if (rv != 0) {
@@ -910,7 +865,6 @@ int dqlite_node_start(dqlite_node *t)
 
 err_after_acquire_dir:
 	release_dir(t->lock_fd);
-err:
 	return rv;
 }
 
