@@ -1,10 +1,9 @@
 #include "heap.h"
+#include "fault.h"
+#include "test/lib/munit.h"
 
 #include <stdlib.h>
 
-#include "fault.h"
-#include "munit.h"
-
 struct heap
 {
     size_t alignment;   /* Value of last aligned alloc */
