@@ -26,16 +26,16 @@ TEST_MODULE(replication_v1);
 		SETUP_LEADER(i);                            \
 	}
 
-#define SETUP_LEADER(I)                                           \
-	do {                                                      \
-		struct leader *leader = &f->leaders[I];           \
-		struct registry *registry = CLUSTER_REGISTRY(I);  \
-		struct db *db;                                    \
-		int rc2;                                          \
-		rc2 = registry__db_get(registry, "test.db", &db); \
-		munit_assert_int(rc2, ==, 0);                     \
-		rc2 = leader__init(leader, db, CLUSTER_RAFT(I));  \
-		munit_assert_int(rc2, ==, 0);                     \
+#define SETUP_LEADER(I)                                                  \
+	do {                                                             \
+		struct leader *leader = &f->leaders[I];                  \
+		struct registry *registry = CLUSTER_REGISTRY(I);         \
+		struct db *db;                                           \
+		int rc2;                                                 \
+		rc2 = registry__get_or_create(registry, "test.db", &db); \
+		munit_assert_int(rc2, ==, 0);                            \
+		rc2 = leader__init(leader, db, CLUSTER_RAFT(I));         \
+		munit_assert_int(rc2, ==, 0);                            \
 	} while (0)
 
 #define TEAR_DOWN                         \
@@ -126,7 +126,7 @@ TEST_MODULE(replication_v1);
 		rv_ = file_->pMethods->xFileSize(file_, &size_);         \
 		munit_assert_int(rv_, ==, 0);                            \
 		pages_ = formatWalCalcFramesNumber(                      \
-		    leader_->db->config->page_size, size_);              \
+		    leader_->db->config->vfs.page_size, size_);          \
 		munit_assert_int(pages_, ==, N);                         \
 	}
 
@@ -185,7 +185,10 @@ struct exec_fixture {
 
 static void fixture_exec_work_cb(struct exec *req)
 {
-	int rv = sqlite3_step(req->stmt);
+	int rv = SQLITE_ROW;
+	while (rv == SQLITE_ROW) {
+		rv = sqlite3_step(req->stmt);
+	}
 	sqlite3_reset(req->stmt);
 
 	if (rv == SQLITE_DONE) {
@@ -312,7 +315,7 @@ TEST_CASE(exec, checkpoint, NULL)
 	struct exec_fixture *f = data;
 	struct config *config = CLUSTER_CONFIG(0);
 	(void)params;
-	config->checkpoint_threshold = 3;
+	config->vfs.checkpoint_threshold = 3;
 	CLUSTER_ELECT(0);
 	EXEC_SQL(0, "CREATE TABLE test (n  INT)");
 	EXEC_SQL(0, "INSERT INTO test(n) VALUES(1)");
@@ -332,13 +335,13 @@ TEST_CASE(exec, checkpoint_read_lock, NULL)
 	char *errmsg;
 	int rv;
 	(void)params;
-	config->checkpoint_threshold = 3;
+	config->vfs.checkpoint_threshold = 3;
 
 	CLUSTER_ELECT(0);
 	EXEC_SQL(0, "CREATE TABLE test (n  INT)");
 
 	/* Initialize another leader. */
-	rv = registry__db_get(registry, "test.db", &db);
+	rv = registry__get_or_create(registry, "test.db", &db);
 	munit_assert_int(rv, ==, 0);
 	leader__init(&leader2, db, CLUSTER_RAFT(0));
 
@@ -404,12 +407,22 @@ static void execCb(struct exec *req)
 	f->status = req->status;
 }
 
+static bool executed(struct raft_fixture *rf, void *unused)
+{
+	(void)unused;
+	struct fixture *f = CONTAINER_OF(rf, struct fixture, cluster);
+	return f->invoked; 
+}
+
 static void fixture_exec(struct fixture *f, unsigned i)
 {
 	f->req.stmt = f->stmt;
+	f->invoked = false;
 	leader_exec(LEADER(i), &f->req, fixture_exec_work_cb, execCb);
+	raft_fixture_step_until(&f->cluster, executed, NULL, 1000);
 }
 
+
 TEST(replication, exec, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
@@ -421,8 +434,6 @@ TEST(replication, exec, setUp, tearDown, 0, NULL)
 	CLUSTER_APPLIED(2);
 	munit_assert_true(f->invoked);
 	munit_assert_int(f->status, ==, RAFT_OK);
-	f->invoked = false;
-	f->status = -1;
 	FINALIZE;
 
 	PREPARE(0, "CREATE TABLE test (a  INT)");
@@ -430,12 +441,10 @@ TEST(replication, exec, setUp, tearDown, 0, NULL)
 	CLUSTER_STEP;
 	munit_assert_true(f->invoked);
 	munit_assert_int(f->status, ==, RAFT_OK);
-	f->invoked = false;
 	FINALIZE;
 
 	PREPARE(0, "COMMIT");
 	fixture_exec(f, 0);
-	munit_assert_false(f->invoked);
 
 	CLUSTER_APPLIED(3);
 	FINALIZE;
@@ -454,13 +463,76 @@ TEST(replication, exec, setUp, tearDown, 0, NULL)
 	return MUNIT_OK;
 }
 
+static void barrierCb(struct raft_barrier *req, int status) {
+	munit_assert_int(status, ==, RAFT_OK);
+	bool *barrier_done = req->data;
+	*barrier_done = true;
+}
+
+static bool barrierDone(struct raft_fixture *f, void *arg) {
+	(void)f;
+	bool *barrier_done = arg;
+	return *barrier_done;
+}
+
+/* Make sure that no barrier is ever issued for a write transaction. */
+TEST(replication, barriers, setUp, tearDown, 0, NULL)
+{
+	struct fixture *f = data;
+
+	/* Make sure work is faster than applying a barrier */
+	raft_fixture_set_work_duration(&f->cluster, 0, 0);
+
+	CLUSTER_ELECT(0);
+
+	/* Right after becoming a leader, the node needs to append a barrier
+	 * before being able to append new items. Make sure that executing a SQL
+	 * statement will not issue another barrier, but will wait for the cluster
+	 * to apply that barrier. */
+	struct raft *r = raft_fixture_get(&f->cluster, 0);
+	int last_index = raft_last_index(r);
+	munit_assert_int(raft_last_applied(r), <, last_index);
+	PREPARE(0, "SELECT 1");
+	EXEC(0);
+	FINALIZE;
+	munit_assert_int(raft_last_index(r), ==, last_index);
+	munit_assert_int(raft_last_applied(r), ==, last_index);
+
+	/* Insert a custom barrier in the log. It is then possible to check if
+	 * any command issues a real barrier because then the custom one will be
+	 * awaited. If the custom barrier didn't complete it means no query
+	 * issued another barrier. */
+	bool barrier_done = false;
+	struct raft_barrier barrier = {
+		.data = &barrier_done,
+	};
+	raft_barrier(CLUSTER_RAFT(0), &barrier, barrierCb);
+
+	/* Read transactions do not need a barrier */
+	PREPARE(0, "SELECT 1");
+	EXEC(0);
+	FINALIZE;
+	munit_assert_false(barrier_done);
+
+	/* Write transactions do not need a barrier, too */
+	PREPARE(0, "BEGIN IMMEDIATE");
+	EXEC(0);
+	FINALIZE;
+	munit_assert_false(barrier_done);
+
+	raft_fixture_step_until(&f->cluster, barrierDone, &barrier_done, 1000);
+
+	return MUNIT_OK;
+}
+
+
 /* If the WAL size grows beyond the configured threshold, checkpoint it. */
 TEST(replication, checkpoint, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	struct config *config = CLUSTER_CONFIG(0);
 
-	config->checkpoint_threshold = 3;
+	config->vfs.checkpoint_threshold = 3;
 
 	CLUSTER_ELECT(0);
 
@@ -471,7 +543,7 @@ TEST(replication, checkpoint, setUp, tearDown, 0, NULL)
 
 	PREPARE(0, "INSERT INTO test(n) VALUES(1)");
 	fixture_exec(f, 0);
-	CLUSTER_APPLIED(5);
+	CLUSTER_APPLIED(4);
 	FINALIZE;
 
 	/* The WAL was truncated. */
@@ -479,3 +551,37 @@ TEST(replication, checkpoint, setUp, tearDown, 0, NULL)
 
 	return MUNIT_OK;
 }
+
+TEST(replication, leaderToFollowerBusy, setUp, tearDown, 0, NULL)
+{
+	struct fixture *f = data;
+
+	CLUSTER_ELECT(0);
+	/* Make sure the leader write-locks the database. */
+	PREPARE(0, "BEGIN IMMEDIATE");
+	fixture_exec(f, 0);
+	FINALIZE;
+
+	/* Lose leadership in the middle of a write transaction */
+	CLUSTER_DEPOSE;
+	SETUP_LEADER(1);
+	CLUSTER_ELECT(1);
+	PREPARE(1, "CREATE TABLE test(i ANY)")
+	fixture_exec(f, 1);
+	FINALIZE;
+	TEAR_DOWN_LEADER(1);
+	CLUSTER_COMMITTED(4);
+	raft_fixture_step_until_applied(&f->cluster, 1, 4, 1000);
+	raft_fixture_step_until_applied(&f->cluster, 2, 4, 1000);
+
+	/* Re-elect the first one and make sure that the table is there */
+	TEAR_DOWN_LEADER(0);
+	CLUSTER_DEPOSE;
+	SETUP_LEADER(0);
+	CLUSTER_ELECT(0);
+	PREPARE(0, "DROP TABLE test");
+	fixture_exec(f, 0);
+	FINALIZE;
+
+	return MUNIT_OK;
+}
\ No newline at end of file
