@@ -5,23 +5,14 @@
 
 #include "../../include/dqlite.h"
 
-#include "../lib/config.h"
 #include "../lib/fs.h"
 #include "../lib/heap.h"
 #include "../lib/runner.h"
 #include "../lib/sqlite.h"
 
 #include "../../src/format.h"
-#include "../../src/raft.h"
 #include "../../src/vfs.h"
 
-static char *bools[] = {"0", "1", NULL};
-
-static MunitParameterEnum vfs_params[] = {
-    {"disk_mode", bools},
-    {NULL, NULL},
-};
-
 /******************************************************************************
  *
  * Fixture
@@ -34,6 +25,7 @@ struct fixture
 	struct sqlite3_vfs vfs;
 	char *dir;
 	char path[VFS_PATH_SZ];
+	struct vfsConfig config;
 };
 
 static void vfsFillPath(struct fixture *f, char *filename)
@@ -49,56 +41,20 @@ static void vfsFillPath(struct fixture *f, char *filename)
 	munit_assert_int(rv, <, VFS_PATH_SZ);
 }
 
-/* Sets the page_size in disk_mode. */
-static void setPageSizeDisk(const MunitParameter params[],
-			    sqlite3_file *f,
-			    unsigned page_size,
-			    int rv)
-{
-	int rc;
-	bool disk_mode = false;
-	char page_sz[32];
-	rc = snprintf(page_sz, sizeof(page_sz), "%u", page_size);
-	munit_assert_int(rc, >, 0);
-	munit_assert_int(rc, <, sizeof(page_sz));
-
-	char *fnctl[] = {
-	    "",
-	    "page_size",
-	    "512",
-	    "",
-	};
-
-	const char *disk_mode_param = munit_parameters_get(params, "disk_mode");
-	if (disk_mode_param != NULL) {
-		disk_mode = (bool)atoi(disk_mode_param);
-	}
-	if (disk_mode) {
-		rc = f->pMethods->xFileControl(f, SQLITE_FCNTL_PRAGMA, fnctl);
-		munit_assert_int(rc, ==, rv);
-	}
-}
-
 static void *setUp(const MunitParameter params[], void *user_data)
 {
 	struct fixture *f = munit_malloc(sizeof *f);
 	int rv;
-	bool disk_mode = false;
 
 	SETUP_HEAP;
 	SETUP_SQLITE;
-	rv = VfsInit(&f->vfs, "dqlite");
+	f->config = (struct vfsConfig) {
+		.name = "dqlite",
+		.page_size = 512,
+	};
+	rv = VfsInit(&f->vfs, &f->config);
 	munit_assert_int(rv, ==, 0);
 	f->dir = NULL;
-	const char *disk_mode_param = munit_parameters_get(params, "disk_mode");
-	if (disk_mode_param != NULL) {
-		disk_mode = (bool)atoi(disk_mode_param);
-	}
-	if (disk_mode) {
-		rv = VfsEnableDisk(&f->vfs);
-		munit_assert_int(rv, ==, 0);
-		f->dir = test_dir_setup();
-	}
 	rv = sqlite3_vfs_register(&f->vfs, 0);
 	munit_assert_int(rv, ==, 0);
 	return f;
@@ -190,30 +146,21 @@ static void *__buf_page_2(void)
 	return buf;
 }
 
-/* Helper to execute a SQL statement. */
-static void __db_exec(sqlite3 *db, const char *sql)
-{
-	int rc;
+#define EXEC(DB, SQL)                                                  \
+	do {                                                           \
+		int exec_rc = sqlite3_exec(DB, SQL, NULL, NULL, NULL); \
+		munit_assert_int(exec_rc, ==, SQLITE_OK);                   \
+	} while (0)
 
-	rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
-	munit_assert_int(rc, ==, SQLITE_OK);
-}
-
-/* Helper to open and initialize a database, setting the page size and
- * WAL mode. */
-static sqlite3 *__db_open(void)
+static sqlite3 *test_db_open(void)
 {
 	sqlite3 *db;
-	int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
+	int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_EXRESCODE;
 	int rc;
 
 	rc = sqlite3_open_v2("test.db", &db, flags, "dqlite");
 	munit_assert_int(rc, ==, SQLITE_OK);
 
-	__db_exec(db, "PRAGMA page_size=512");
-	__db_exec(db, "PRAGMA synchronous=OFF");
-	__db_exec(db, "PRAGMA journal_mode=WAL");
-
 	return db;
 }
 
@@ -229,15 +176,16 @@ static void __db_close(sqlite3 *db)
  * given database. */
 static uint32_t __wal_idx_mx_frame(sqlite3 *db)
 {
+	const int WAL_INDEX_REGION_SIZE = 32768;
 	sqlite3_file *file;
 	volatile void *region;
 	uint32_t mx_frame;
 	int rc;
 
-	rc = sqlite3_file_control(db, "main", SQLITE_FCNTL_FILE_POINTER, &file);
+	rc = sqlite3_file_control(db, NULL, SQLITE_FCNTL_FILE_POINTER, &file);
 	munit_assert_int(rc, ==, SQLITE_OK);
 
-	rc = file->pMethods->xShmMap(file, 0, 0, 0, &region);
+	rc = file->pMethods->xShmMap(file, 0, WAL_INDEX_REGION_SIZE, 0, &region);
 	munit_assert_int(rc, ==, SQLITE_OK);
 
 	/* The mxFrame number is 16th byte of the WAL index header. See also
@@ -259,10 +207,10 @@ static uint32_t *__wal_idx_read_marks(sqlite3 *db)
 
 	marks = munit_malloc(FORMAT__WAL_NREADER * sizeof *marks);
 
-	rc = sqlite3_file_control(db, "main", SQLITE_FCNTL_FILE_POINTER, &file);
+	rc = sqlite3_file_control(db, NULL, SQLITE_FCNTL_FILE_POINTER, &file);
 	munit_assert_int(rc, ==, SQLITE_OK);
 
-	rc = file->pMethods->xShmMap(file, 0, 0, 0, &region);
+	rc = file->pMethods->xShmMap(file, 0, 32768, 0, &region);
 	munit_assert_int(rc, ==, SQLITE_OK);
 
 	/* The read-mark array starts at the 100th byte of the WAL index
@@ -282,7 +230,7 @@ static int __shm_shared_lock_held(sqlite3 *db, int i)
 	int locked;
 	int rc;
 
-	rc = sqlite3_file_control(db, "main", SQLITE_FCNTL_FILE_POINTER, &file);
+	rc = sqlite3_file_control(db, NULL, SQLITE_FCNTL_FILE_POINTER, &file);
 	munit_assert_int(rc, ==, SQLITE_OK);
 
 	/* Try to acquire an exclusive lock, which will fail if the shared lock
@@ -311,7 +259,7 @@ SUITE(VfsOpen)
 
 /* If the EXCLUSIVE and CREATE flag are given, and the file already exists, an
  * error is returned. */
-TEST(VfsOpen, exclusive, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, exclusive, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file1 = munit_malloc(f->vfs.szOsFile);
@@ -345,7 +293,7 @@ TEST(VfsOpen, exclusive, setUp, tearDown, 0, vfs_params)
 
 /* It's possible to open again a previously created file. In that case passing
  * SQLITE_OPEN_CREATE is not necessary. */
-TEST(VfsOpen, again, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, again, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -379,7 +327,7 @@ TEST(VfsOpen, again, setUp, tearDown, 0, vfs_params)
 
 /* If the file does not exist and the SQLITE_OPEN_CREATE flag is not passed, an
  * error is returned. */
-TEST(VfsOpen, noent, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, noent, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -402,7 +350,7 @@ TEST(VfsOpen, noent, setUp, tearDown, 0, vfs_params)
 
 /* Trying to open a WAL file before its main database file results in an
  * error. */
-TEST(VfsOpen, walBeforeDb, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, walBeforeDb, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -423,13 +371,12 @@ TEST(VfsOpen, walBeforeDb, setUp, tearDown, 0, vfs_params)
 	return MUNIT_OK;
 }
 
-/* Trying to run queries against a database that hasn't turned off the
- * synchronous flag results in an error. */
-TEST(VfsOpen, synchronous, setUp, tearDown, 0, vfs_params)
+/* Opening a connection also sets synchronous to OFF */
+TEST(VfsOpen, synchronous, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3 *db;
-	int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE;
+	int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_EXRESCODE;
 	int rc;
 
 	(void)params;
@@ -441,12 +388,18 @@ TEST(VfsOpen, synchronous, setUp, tearDown, 0, vfs_params)
 	rc = sqlite3_open_v2(f->path, &db, flags, f->vfs.zName);
 	munit_assert_int(rc, ==, SQLITE_OK);
 
-	__db_exec(db, "PRAGMA page_size=4092");
+	sqlite3_stmt *stmt;
+	rc = sqlite3_prepare_v2(db, "PRAGMA synchronous", -1, &stmt, NULL);
+	munit_assert_int(rc, ==, SQLITE_OK);
 
-	rc = sqlite3_exec(db, "PRAGMA journal_mode=WAL", NULL, NULL, NULL);
-	munit_assert_int(rc, ==, SQLITE_IOERR);
+	rc = sqlite3_step(stmt);
+	munit_assert_int(rc, ==, SQLITE_ROW);
+	munit_assert_int(sqlite3_column_int(stmt, 0), ==, 0);
 
-	munit_assert_string_equal(sqlite3_errmsg(db), "disk I/O error");
+	rc = sqlite3_step(stmt);
+	munit_assert_int(rc, ==, SQLITE_DONE);
+
+	sqlite3_finalize(stmt);
 
 	__db_close(db);
 
@@ -457,7 +410,7 @@ TEST(VfsOpen, synchronous, setUp, tearDown, 0, vfs_params)
 }
 
 /* Out of memory when creating the content structure for a new file. */
-TEST(VfsOpen, oom, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, oom, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -479,7 +432,7 @@ TEST(VfsOpen, oom, setUp, tearDown, 0, vfs_params)
 }
 
 /* Out of memory when internally copying the filename. */
-TEST(VfsOpen, oomFilename, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, oomFilename, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -501,20 +454,17 @@ TEST(VfsOpen, oomFilename, setUp, tearDown, 0, vfs_params)
 }
 
 /* Open a temporary file. */
-TEST(VfsOpen, tmp, setUp, tearDown, 0, vfs_params)
+TEST(VfsOpen, tmp, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
-	int flags = 0;
 	char buf[16];
 	int rc;
 
 	(void)params;
 
-	flags |= SQLITE_OPEN_CREATE;
-	flags |= SQLITE_OPEN_READWRITE;
-	flags |= SQLITE_OPEN_TEMP_JOURNAL;
-	flags |= SQLITE_OPEN_DELETEONCLOSE;
+	int flags = SQLITE_OPEN_CREATE | SQLITE_OPEN_READWRITE |
+		    SQLITE_OPEN_TEMP_JOURNAL | SQLITE_OPEN_DELETEONCLOSE;
 
 	rc = f->vfs.xOpen(&f->vfs, NULL, file, flags, &flags);
 	munit_assert_int(rc, ==, SQLITE_OK);
@@ -545,7 +495,7 @@ TEST(VfsOpen, tmp, setUp, tearDown, 0, vfs_params)
 SUITE(VfsDelete)
 
 /* Delete a file. */
-TEST(VfsDelete, success, setUp, tearDown, 0, vfs_params)
+TEST(VfsDelete, success, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -555,7 +505,7 @@ TEST(VfsDelete, success, setUp, tearDown, 0, vfs_params)
 
 	(void)params;
 
-	vfsFillPath(f, "test.db");
+	vfsFillPath(f, "test.db-wal");
 	rc = f->vfs.xOpen(&f->vfs, f->path, file, flags, &flags);
 	munit_assert_int(rc, ==, 0);
 
@@ -575,23 +525,6 @@ TEST(VfsDelete, success, setUp, tearDown, 0, vfs_params)
 	return MUNIT_OK;
 }
 
-/* Trying to delete a non-existing file results in an error. */
-TEST(VfsDelete, enoent, setUp, tearDown, 0, vfs_params)
-{
-	struct fixture *f = data;
-
-	int rc;
-
-	(void)params;
-
-	vfsFillPath(f, "test.db");
-	rc = f->vfs.xDelete(&f->vfs, f->path, 0);
-	munit_assert_int(rc, ==, SQLITE_IOERR_DELETE_NOENT);
-	munit_assert_int(ENOENT, ==, f->vfs.xGetLastError(&f->vfs, 0, 0));
-
-	return MUNIT_OK;
-}
-
 /******************************************************************************
  *
  * xAccess
@@ -601,7 +534,7 @@ TEST(VfsDelete, enoent, setUp, tearDown, 0, vfs_params)
 SUITE(VfsAccess)
 
 /* Accessing an existing file returns true. */
-TEST(VfsAccess, success, setUp, tearDown, 0, vfs_params)
+TEST(VfsAccess, success, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -615,7 +548,6 @@ TEST(VfsAccess, success, setUp, tearDown, 0, vfs_params)
 	rc = f->vfs.xOpen(&f->vfs, f->path, file, flags, &flags);
 	munit_assert_int(rc, ==, 0);
 
-	setPageSizeDisk(params, file, 512, SQLITE_NOTFOUND);
 	/* Write the first page, containing the header and some content. */
 	void *buf_page_1 = __buf_page_1();
 	rc = file->pMethods->xWrite(file, buf_page_1, 512, 0);
@@ -635,7 +567,7 @@ TEST(VfsAccess, success, setUp, tearDown, 0, vfs_params)
 }
 
 /* Trying to access a non existing file returns false. */
-TEST(VfsAccess, noent, setUp, tearDown, 0, vfs_params)
+TEST(VfsAccess, noent, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 
@@ -662,7 +594,7 @@ TEST(VfsAccess, noent, setUp, tearDown, 0, vfs_params)
 SUITE(VfsFullPathname);
 
 /* The xFullPathname API returns the filename unchanged. */
-TEST(VfsFullPathname, success, setUp, tearDown, 0, vfs_params)
+TEST(VfsFullPathname, success, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 
@@ -679,40 +611,6 @@ TEST(VfsFullPathname, success, setUp, tearDown, 0, vfs_params)
 	return MUNIT_OK;
 }
 
-/******************************************************************************
- *
- * xClose
- *
- ******************************************************************************/
-
-SUITE(VfsClose)
-
-/* Closing a file decreases its refcount so it's possible to delete it. */
-TEST(VfsClose, thenDelete, setUp, tearDown, 0, vfs_params)
-{
-	struct fixture *f = data;
-	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
-
-	int flags = SQLITE_OPEN_CREATE | SQLITE_OPEN_MAIN_DB;
-	int rc;
-
-	(void)params;
-	vfsFillPath(f, "test.db");
-
-	rc = f->vfs.xOpen(&f->vfs, f->path, file, flags, &flags);
-	munit_assert_int(rc, ==, 0);
-
-	rc = file->pMethods->xClose(file);
-	munit_assert_int(rc, ==, 0);
-
-	rc = f->vfs.xDelete(&f->vfs, f->path, 0);
-	munit_assert_int(rc, ==, 0);
-
-	free(file);
-
-	return MUNIT_OK;
-}
-
 /******************************************************************************
  *
  * xRead
@@ -722,7 +620,7 @@ TEST(VfsClose, thenDelete, setUp, tearDown, 0, vfs_params)
 SUITE(VfsRead)
 
 /* Trying to read a file that was not written yet, results in an error. */
-TEST(VfsRead, neverWritten, setUp, tearDown, 0, vfs_params)
+TEST(VfsRead, neverWritten, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = __file_create_main_db(f);
@@ -755,7 +653,7 @@ TEST(VfsRead, neverWritten, setUp, tearDown, 0, vfs_params)
 SUITE(VfsWrite)
 
 /* Write the header of the database file. */
-TEST(VfsWrite, dbHeader, setUp, tearDown, 0, vfs_params)
+TEST(VfsWrite, dbHeader, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = __file_create_main_db(f);
@@ -780,7 +678,7 @@ TEST(VfsWrite, dbHeader, setUp, tearDown, 0, vfs_params)
 
 /* Write the header of the database file, then the full first page and a second
  * page. */
-TEST(VfsWrite, andReadPages, setUp, tearDown, 0, vfs_params)
+TEST(VfsWrite, andReadPages, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = __file_create_main_db(f);
@@ -803,9 +701,6 @@ TEST(VfsWrite, andReadPages, setUp, tearDown, 0, vfs_params)
 	rc = file->pMethods->xWrite(file, buf_page_1, 512, 0);
 	munit_assert_int(rc, ==, 0);
 
-	/* Set the page_size in disk_mode */
-	setPageSizeDisk(params, file, 512, SQLITE_NOTFOUND);
-
 	/* Write a second page. */
 	rc = file->pMethods->xWrite(file, buf_page_2, 512, 512);
 	munit_assert_int(rc, ==, 0);
@@ -948,6 +843,41 @@ TEST(VfsWrite, beyondLast, setUp, tearDown, 0, NULL)
 	return MUNIT_OK;
 }
 
+/******************************************************************************
+ *
+ * xFileControl
+ *
+ ******************************************************************************/
+SUITE(VfsFileControl)
+
+TEST(VfsFileControl, persistWal, setUp, tearDown, 0, NULL)
+{
+	struct fixture *f = data;
+	sqlite3_file *file = __file_create_main_db(f);
+
+	/* Setting to false should fail. */
+	int value = 0;
+	int rv = file->pMethods->xFileControl(file, SQLITE_FCNTL_PERSIST_WAL, &value);
+	munit_assert_int(rv, ==, SQLITE_MISUSE);
+	munit_assert_int(value, ==, 0);
+
+	/* Setting to true is ok. */
+	value = 1;
+	rv = file->pMethods->xFileControl(file, SQLITE_FCNTL_PERSIST_WAL, &value);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	munit_assert_int(value, ==, 1);
+
+	/* Querying the value is ok. */
+	value = -1;
+	rv = file->pMethods->xFileControl(file, SQLITE_FCNTL_PERSIST_WAL, &value);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	munit_assert_int(value, ==, 1);
+
+	free(file);
+
+	return MUNIT_OK;
+}
+
 /******************************************************************************
  *
  * xTruncate
@@ -957,7 +887,7 @@ TEST(VfsWrite, beyondLast, setUp, tearDown, 0, NULL)
 SUITE(VfsTruncate);
 
 /* Truncate the main database file. */
-TEST(VfsTruncate, database, setUp, tearDown, 0, vfs_params)
+TEST(VfsTruncate, database, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = __file_create_main_db(f);
@@ -984,9 +914,6 @@ TEST(VfsTruncate, database, setUp, tearDown, 0, vfs_params)
 	munit_assert_int(rc, ==, 0);
 	munit_assert_int(size, ==, 0);
 
-	/* Set the page_size in disk_mode */
-	setPageSizeDisk(params, file, 512, SQLITE_NOTFOUND);
-
 	/* Write the first page, containing the header. */
 	rc = file->pMethods->xWrite(file, buf_page_1, 512, 0);
 	munit_assert_int(rc, ==, 0);
@@ -1029,7 +956,7 @@ TEST(VfsTruncate, database, setUp, tearDown, 0, vfs_params)
 
 /* Truncating a file which is not the main db file or the WAL file produces an
  * error. */
-TEST(VfsTruncate, unexpected, setUp, tearDown, 0, vfs_params)
+TEST(VfsTruncate, unexpected, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *main_db = __file_create_main_db(f);
@@ -1056,7 +983,7 @@ TEST(VfsTruncate, unexpected, setUp, tearDown, 0, vfs_params)
 	rc = file->pMethods->xTruncate(file, 0);
 	munit_assert_int(rc, ==, SQLITE_IOERR_TRUNCATE);
 
-	rc = file->pMethods->xClose(main_db);
+	rc = main_db->pMethods->xClose(main_db);
 	munit_assert_int(rc, ==, 0);
 	free(main_db);
 
@@ -1068,7 +995,7 @@ TEST(VfsTruncate, unexpected, setUp, tearDown, 0, vfs_params)
 }
 
 /* Truncating an empty file is a no-op. */
-TEST(VfsTruncate, empty, setUp, tearDown, 0, vfs_params)
+TEST(VfsTruncate, empty, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = __file_create_main_db(f);
@@ -1149,7 +1076,7 @@ TEST(VfsTruncate, misaligned, setUp, tearDown, 0, NULL)
 
 SUITE(VfsShmMap);
 
-static char *test_shm_map_oom_delay[] = {"0", "1", NULL};
+static char *test_shm_map_oom_delay[] = {"0", NULL};
 static char *test_shm_map_oom_repeat[] = {"1", NULL};
 
 static MunitParameterEnum test_shm_map_oom_params[] = {
@@ -1189,7 +1116,7 @@ SUITE(VfsShmLock)
 
 /* If an exclusive lock is in place, getting a shared lock on any index of its
  * range fails. */
-TEST(VfsShmLock, sharedBusy, setUp, tearDown, 0, vfs_params)
+TEST(VfsShmLock, sharedBusy, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -1217,6 +1144,9 @@ TEST(VfsShmLock, sharedBusy, setUp, tearDown, 0, vfs_params)
 	rc = file->pMethods->xShmLock(file, 3, 1, flags);
 	munit_assert_int(rc, ==, SQLITE_BUSY);
 
+	rc = file->pMethods->xShmUnmap(file, 0);
+	munit_assert_int(rc, ==, 0);
+
 	rc = file->pMethods->xClose(file);
 	munit_assert_int(rc, ==, 0);
 	free(file);
@@ -1226,10 +1156,11 @@ TEST(VfsShmLock, sharedBusy, setUp, tearDown, 0, vfs_params)
 
 /* If a shared lock is in place on any of the indexes of the requested range,
  * getting an exclusive lock fails. */
-TEST(VfsShmLock, exclBusy, setUp, tearDown, 0, vfs_params)
+TEST(VfsShmLock, exclBusy, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
-	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
+	sqlite3_file *file1 = munit_malloc(f->vfs.szOsFile);
+	sqlite3_file *file2 = munit_malloc(f->vfs.szOsFile);
 	int flags = SQLITE_OPEN_CREATE | SQLITE_OPEN_MAIN_DB;
 	volatile void *region;
 	int rc;
@@ -1239,33 +1170,43 @@ TEST(VfsShmLock, exclBusy, setUp, tearDown, 0, vfs_params)
 
 	vfsFillPath(f, "test.db");
 
-	rc = f->vfs.xOpen(&f->vfs, f->path, file, flags, &flags);
+	rc = f->vfs.xOpen(&f->vfs, f->path, file1, flags, &flags);
 	munit_assert_int(rc, ==, 0);
 
-	rc = file->pMethods->xShmMap(file, 0, 32768, 1, &region);
+	rc = f->vfs.xOpen(&f->vfs, f->path, file2, flags, &flags);
+	munit_assert_int(rc, ==, 0);
+
+	rc = file1->pMethods->xShmMap(file1, 0, 32768, 1, &region);
 	munit_assert_int(rc, ==, 0);
 
 	/* Take a shared lock on index 3. */
 	flags = SQLITE_SHM_LOCK | SQLITE_SHM_SHARED;
-	rc = file->pMethods->xShmLock(file, 3, 1, flags);
+	rc = file1->pMethods->xShmLock(file1, 3, 1, flags);
 	munit_assert_int(rc, ==, 0);
 
 	/* Attempting to get an exclusive lock on a range that contains index 3
 	 * fails. */
 	flags = SQLITE_SHM_LOCK | SQLITE_SHM_EXCLUSIVE;
-	rc = file->pMethods->xShmLock(file, 2, 3, flags);
+	rc = file2->pMethods->xShmLock(file2, 2, 3, flags);
 	munit_assert_int(rc, ==, SQLITE_BUSY);
 
-	rc = file->pMethods->xClose(file);
+	rc = file1->pMethods->xShmUnmap(file1, 0);
 	munit_assert_int(rc, ==, 0);
-	free(file);
+
+	rc = file2->pMethods->xClose(file2);
+	munit_assert_int(rc, ==, 0);
+	free(file2);
+
+	rc = file1->pMethods->xClose(file1);
+	munit_assert_int(rc, ==, 0);
+	free(file1);
 
 	return MUNIT_OK;
 }
 
 /* The native unix VFS implementation from SQLite allows to release a shared
  * memory lock without acquiring it first. */
-TEST(VfsShmLock, releaseUnix, setUp, tearDown, 0, vfs_params)
+TEST(VfsShmLock, releaseUnix, setUp, tearDown, 0, NULL)
 {
 	(void)data;
 	struct sqlite3_vfs *vfs = sqlite3_vfs_find("unix");
@@ -1320,7 +1261,7 @@ TEST(VfsShmLock, releaseUnix, setUp, tearDown, 0, vfs_params)
 /* The dqlite VFS implementation allows to release a shared memory lock without
  * acquiring it first. This is important because at open time sometimes SQLite
  * will do just that (release before acquire). */
-TEST(VfsShmLock, release, setUp, tearDown, 0, vfs_params)
+TEST(VfsShmLock, release, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	sqlite3_file *file = munit_malloc(f->vfs.szOsFile);
@@ -1357,46 +1298,6 @@ TEST(VfsShmLock, release, setUp, tearDown, 0, vfs_params)
 	return MUNIT_OK;
 }
 
-/******************************************************************************
- *
- * xFileControl
- *
- ******************************************************************************/
-
-SUITE(VfsFileControl)
-
-/* Trying to set the journal mode to anything other than "wal" produces an
- * error. */
-TEST(VfsFileControl, journal, setUp, tearDown, 0, vfs_params)
-{
-	struct fixture *f = data;
-	sqlite3_file *file = __file_create_main_db(f);
-	char *fnctl[] = {
-	    "",
-	    "journal_mode",
-	    "memory",
-	    "",
-	};
-	int rc;
-
-	(void)params;
-	(void)data;
-
-	/* Setting the page size a first time returns NOTFOUND, which is what
-	 * SQLite effectively expects. */
-	rc = file->pMethods->xFileControl(file, SQLITE_FCNTL_PRAGMA, fnctl);
-	munit_assert_int(rc, ==, SQLITE_IOERR);
-
-	rc = file->pMethods->xClose(file);
-	munit_assert_int(rc, ==, 0);
-	free(file);
-
-	/* Free allocated memory from call to sqlite3_mprintf */
-	sqlite3_free(fnctl[0]);
-
-	return MUNIT_OK;
-}
-
 /******************************************************************************
  *
  * xCurrentTime
@@ -1405,7 +1306,7 @@ TEST(VfsFileControl, journal, setUp, tearDown, 0, vfs_params)
 
 SUITE(VfsCurrentTime)
 
-TEST(VfsCurrentTime, success, setUp, tearDown, 0, vfs_params)
+TEST(VfsCurrentTime, success, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	double now;
@@ -1430,7 +1331,7 @@ TEST(VfsCurrentTime, success, setUp, tearDown, 0, vfs_params)
 SUITE(VfsSleep)
 
 /* The xSleep implementation is a no-op. */
-TEST(VfsSleep, success, setUp, tearDown, 0, vfs_params)
+TEST(VfsSleep, success, setUp, tearDown, 0, NULL)
 {
 	struct fixture *f = data;
 	int microseconds;
@@ -1463,15 +1364,16 @@ static MunitParameterEnum test_create_oom_params[] = {
 
 TEST(VfsInit, oom, setUp, tearDown, 0, test_create_oom_params)
 {
-	struct sqlite3_vfs vfs;
-	int rv;
-
 	(void)params;
 	(void)data;
 
 	test_heap_fault_enable();
-
-	rv = VfsInit(&vfs, "dqlite");
+	struct vfsConfig conf = {
+		.name = "dqlite",
+		.page_size = 512,
+	};
+	struct sqlite3_vfs vfs;
+	int rv = VfsInit(&vfs, &conf);
 	munit_assert_int(rv, ==, DQLITE_NOMEM);
 
 	return MUNIT_OK;
@@ -1486,7 +1388,7 @@ TEST(VfsInit, oom, setUp, tearDown, 0, test_create_oom_params)
 SUITE(VfsIntegration)
 
 /* Test our expections on the memory-mapped WAl index format. */
-TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
+TEST(VfsIntegration, wal, setUp, tearDown, 0, NULL)
 {
 	sqlite3 *db1;
 	sqlite3 *db2;
@@ -1498,10 +1400,10 @@ TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
 
 	return MUNIT_SKIP;
 
-	db1 = __db_open();
-	db2 = __db_open();
+	db1 = test_db_open();
+	db2 = test_db_open();
 
-	__db_exec(db1, "CREATE TABLE test (n INT)");
+	EXEC(db1, "CREATE TABLE test (n INT)");
 
 	munit_assert_int(__wal_idx_mx_frame(db1), ==, 2);
 
@@ -1514,8 +1416,8 @@ TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
 	free(read_marks);
 
 	/* Start a read transaction on db2 */
-	__db_exec(db2, "BEGIN");
-	__db_exec(db2, "SELECT * FROM test");
+	EXEC(db2, "BEGIN");
+	EXEC(db2, "SELECT * FROM test");
 
 	/* The max frame is set to 2, which is the current size of the WAL. */
 	munit_assert_int(__wal_idx_mx_frame(db2), ==, 2);
@@ -1534,10 +1436,10 @@ TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
 	munit_assert_true(__shm_shared_lock_held(db2, 3 + 1));
 
 	/* Start a write transaction on db1 */
-	__db_exec(db1, "BEGIN");
+	EXEC(db1, "BEGIN");
 
 	for (i = 0; i < 100; i++) {
-		__db_exec(db1, "INSERT INTO test(n) VALUES(1)");
+		EXEC(db1, "INSERT INTO test(n) VALUES(1)");
 	}
 
 	/* The mx frame is still 2 since the transaction is not committed. */
@@ -1552,7 +1454,7 @@ TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
 	munit_assert_uint32(read_marks[4], ==, 0xffffffff);
 	free(read_marks);
 
-	__db_exec(db1, "COMMIT");
+	EXEC(db1, "COMMIT");
 
 	/* The mx frame is now 6. */
 	munit_assert_int(__wal_idx_mx_frame(db1), ==, 6);
@@ -1561,8 +1463,8 @@ TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
 	munit_assert_true(__shm_shared_lock_held(db2, 3 + 1));
 
 	/* Start a read transaction on db1 */
-	__db_exec(db1, "BEGIN");
-	__db_exec(db1, "SELECT * FROM test");
+	EXEC(db1, "BEGIN");
+	EXEC(db1, "SELECT * FROM test");
 
 	/* The mx frame is still unchanged. */
 	munit_assert_int(__wal_idx_mx_frame(db1), ==, 6);
@@ -1589,7 +1491,7 @@ TEST(VfsIntegration, wal, setUp, tearDown, 0, vfs_params)
 }
 
 /* Full checkpoints are possible only when no read mark is set. */
-TEST(VfsIntegration, checkpoint, setUp, tearDown, 0, vfs_params)
+TEST(VfsIntegration, checkpoint, setUp, tearDown, 0, NULL)
 {
 	sqlite3 *db1;
 	sqlite3 *db2;
@@ -1608,26 +1510,25 @@ TEST(VfsIntegration, checkpoint, setUp, tearDown, 0, vfs_params)
 
 	return MUNIT_SKIP;
 
-	db1 = __db_open();
+	db1 = test_db_open();
 
-	__db_exec(db1, "CREATE TABLE test (n INT)");
+	EXEC(db1, "CREATE TABLE test (n INT)");
 
 	/* Insert a few rows so we grow the size of the WAL. */
-	__db_exec(db1, "BEGIN");
+	EXEC(db1, "BEGIN");
 
 	for (i = 0; i < 500; i++) {
 		sprintf(stmt, "INSERT INTO test(n) VALUES(%d)", i);
-		__db_exec(db1, stmt);
+		EXEC(db1, stmt);
 	}
 
-	__db_exec(db1, "COMMIT");
+	EXEC(db1, "COMMIT");
 
 	/* Get the file objects for the main database and the WAL. */
-	rv = sqlite3_file_control(db1, "main", SQLITE_FCNTL_FILE_POINTER,
-				  &file1);
+	rv = sqlite3_file_control(db1, NULL, SQLITE_FCNTL_FILE_POINTER, &file1);
 	munit_assert_int(rv, ==, 0);
 
-	rv = sqlite3_file_control(db1, "main", SQLITE_FCNTL_JOURNAL_POINTER,
+	rv = sqlite3_file_control(db1, NULL, SQLITE_FCNTL_JOURNAL_POINTER,
 				  &file2);
 	munit_assert_int(rv, ==, 0);
 
@@ -1640,9 +1541,9 @@ TEST(VfsIntegration, checkpoint, setUp, tearDown, 0, vfs_params)
 
 	/* Start a read transaction on a different connection, acquiring a
 	 * shared lock on all WAL pages. */
-	db2 = __db_open();
-	__db_exec(db2, "BEGIN");
-	__db_exec(db2, "SELECT * FROM test");
+	db2 = test_db_open();
+	EXEC(db2, "BEGIN");
+	EXEC(db2, "SELECT * FROM test");
 
 	read_marks = __wal_idx_read_marks(db1);
 	munit_assert_int(read_marks[1], ==, 13);
@@ -1656,26 +1557,26 @@ TEST(VfsIntegration, checkpoint, setUp, tearDown, 0, vfs_params)
 
 	/* Execute a new write transaction, deleting some of the pages we
 	 * inserted and creating new ones. */
-	__db_exec(db1, "BEGIN");
-	__db_exec(db1, "DELETE FROM test WHERE n > 200");
+	EXEC(db1, "BEGIN");
+	EXEC(db1, "DELETE FROM test WHERE n > 200");
 
 	for (i = 0; i < 1000; i++) {
 		sprintf(stmt, "INSERT INTO test(n) VALUES(%d)", i);
-		__db_exec(db1, stmt);
+		EXEC(db1, stmt);
 	}
 
-	__db_exec(db1, "COMMIT");
+	EXEC(db1, "COMMIT");
 
 	/* Since there's a shared read lock, a full checkpoint will fail. */
-	rv = sqlite3_wal_checkpoint_v2(db1, "main", SQLITE_CHECKPOINT_TRUNCATE,
+	rv = sqlite3_wal_checkpoint_v2(db1, NULL, SQLITE_CHECKPOINT_TRUNCATE,
 				       &log, &ckpt);
 	munit_assert_int(rv, !=, 0);
 
 	/* If we complete the read transaction the shared lock is realeased and
 	 * the checkpoint succeeds. */
-	__db_exec(db2, "COMMIT");
+	EXEC(db2, "COMMIT");
 
-	rv = sqlite3_wal_checkpoint_v2(db1, "main", SQLITE_CHECKPOINT_TRUNCATE,
+	rv = sqlite3_wal_checkpoint_v2(db1, NULL, SQLITE_CHECKPOINT_TRUNCATE,
 				       &log, &ckpt);
 	munit_assert_int(rv, ==, 0);
 
@@ -1684,3 +1585,29 @@ TEST(VfsIntegration, checkpoint, setUp, tearDown, 0, vfs_params)
 
 	return SQLITE_OK;
 }
+
+/* Check that the VFS does not interfere with normal sqlite execution */
+TEST(VfsIntegration, sqlite, setUp, tearDown, 0, NULL)
+{
+	(void)data;
+	(void)params;
+
+	const int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_EXRESCODE | SQLITE_OPEN_DELETEONCLOSE;
+
+	const char *filenames[] = {
+		"/tmp/dqlite-test-normal-sqlite",
+		"", /* temp db */
+		":memory:", /* in-memory db*/
+		NULL,
+	};
+
+	for (int i = 0; filenames[i]; i++) {
+		sqlite3 *db;
+		int rc = sqlite3_open_v2(filenames[i], &db, flags, NULL);
+		munit_assert_int(rc, ==, SQLITE_OK);
+		EXEC(db, "CREATE TABLE asd(i)");
+		sqlite3_close(db);
+	}
+
+	return SQLITE_OK;
+}
